import json
import logging
import math
from typing import Any

from shapely.geometry import Point

from pymammotion.data.model.hash_list import (
    AreaHashNameList,
    CommDataCouple,
    FrameList,
    HashList,
    MowPath,
    NavGetCommData,
)

logger = logging.getLogger(__name__)


GEOMETRY_TYPES: list[str] = ["Polygon", "Polygon", "LineString", "Point"]
MAP_OBJECT_TYPES: list[str] = ["area", "path", "obstacle", "dump"]

image_path = "/local/community/ha-mammotion-map/dist/assets/map/"

RTK_IMAGE = {
    "iconImage": "map_icon_base_station_rtk.webp",
    "iconSize": [30, 30],
    "iconAnchor": [15, 30],
    "iconUrl": f"{image_path}map_icon_base_station_rtk.webp",
}

DOCK_IMAGE = {
    "iconImage": "icon_map_recharge.webp",
    "iconSize": [30, 30],
    "iconAnchor": [15, 15],
    "iconUrl": f"{image_path}icon_map_recharge.webp",
}

#############################
# STYLE CONFIGURATION
#############################

DOCK_STYLE = {"color": "lightgray", "fill": "lightgray", "weight": 2, "opacity": 1.0, "fillOpacity": 0.7, "radius": 10}

RTK_STYLE = {"color": "purple", "fill": "purple", "weight": 2, "opacity": 1.0, "fillOpacity": 0.7, "radius": 7}

AREA_STYLE = {
    "color": "green",
    "fill": "darkgreen",
    "weight": 3,
    "opacity": 0.8,
    "fillOpacity": 0.3,
    "dashArray": "",
    "lineCap": "round",
    "lineJoin": "round",
}

OBSTACLE_STYLE = {
    "color": "#FF4D00",
    "fill": "darkorange",
    "weight": 2,
    "opacity": 0.9,
    "fillOpacity": 0.4,
    "dashArray": "",
    "lineCap": "round",
    "lineJoin": "round",
}

PATH_STYLE = {
    "color": "#ffffff",
    "weight": 8,
    "opacity": 1.0,
    "zIndex": -1,
    "dashArray": "",
    "lineCap": "round",
    "lineJoin": "round",
    "road_center_color": "#696969",
    "road_center_dash": "8, 8",
}

POINT_STYLE = {"color": "blue", "fill": "lightblue", "weight": 2, "opacity": 1.0, "fillOpacity": 0.7, "radius": 5}

geojson_metadata = {"name": "Lawn Areas", "description": "Generated from Mammotion diagnostics data"}

# Map type IDs
TYPE_MOWING_ZONE: int = 0
TYPE_OBSTACLE: int = 1
TYPE_PATH: int = 2
TYPE_MOW_PATH: int = 4

# Coordinate conversion constants
METERS_PER_DEGREE: int = 111320

# Type aliases
Coordinate = tuple[float, float]
CoordinateList = list[list[float]]
GeoJSONFeature = dict[str, Any]
GeoJSONCollection = dict[str, Any]
LocationDict = dict[str, Coordinate]

# Path to mammotion integration diagnostics file
geometry_types: list[str] = GEOMETRY_TYPES


class GeojsonGenerator:
    """Class for logging GeoJSON data."""

    @staticmethod
    def is_overlapping(p: Point, placed_points: list[Point], min_distance: float = 0.00005) -> bool:
        """Check if point p is too close to any previously placed label."""
        return any(p.distance(existing) < min_distance for existing in placed_points)

    @staticmethod
    def apply_meter_offsets(lon: float, lat: float, lon_offset: float, lat_offset: float) -> list[float]:
        """Apply meter-based offsets to coordinates (in degrees)"""
        new_lon = lon + (lon_offset / (METERS_PER_DEGREE * math.cos(math.radians(lat))))
        new_lat = lat + (lat_offset / METERS_PER_DEGREE)
        return [new_lon, new_lat]

    @staticmethod
    def generate_geojson(
        hash_list: HashList, rtk_location: Point, dock_location: Point, dock_rotation: int
    ) -> GeoJSONCollection:
        """Generate GeoJSON from hash list data.

        Args:
            hash_list: HashList object containing map data
            rtk_location: Tuple of (longitude, latitude) for rtk position
            :param hash_list:
            :param rtk_location:
            :param dock_rotation:
            :param dock_location:

        """
        area_names = GeojsonGenerator._build_area_name_lookup(hash_list.area_name)

        geo_json: GeoJSONCollection = {"type": "FeatureCollection", "name": "Lawn Areas", "features": []}
        GeojsonGenerator._add_rtk_and_dock(rtk_location, dock_location, dock_rotation, geo_json)
        total_frames = GeojsonGenerator._process_map_objects(hash_list, rtk_location, area_names, geo_json)

        # _save_geojson(geo_json)
        return geo_json

    @staticmethod
    def generate_mow_path_geojson(hash_list: HashList, rtk_location: Point) -> GeoJSONCollection:
        """Generate GeoJSON from hash list data."""
        geo_json: GeoJSONCollection = {"type": "FeatureCollection", "name": "Mowing Lawn Areas", "features": []}

        total_frames = GeojsonGenerator._process_mow_map_objects(hash_list, rtk_location, geo_json)
        return geo_json

    @staticmethod
    def _add_rtk_and_dock(
        rtk_location: Point, dock_location: Point, dock_rotation: int, geo_json: GeoJSONCollection
    ) -> None:
        geo_json["features"].append(
            {
                "type": "Feature",
                "properties": {
                    "title": "RTK Base",
                    "Name": "RTK Base",
                    "description": "RTK Base Station location",
                    "type_name": "station",
                    **RTK_STYLE,
                    **RTK_IMAGE,
                },
                "geometry": {"type": "Point", "coordinates": [rtk_location.y, rtk_location.x]},
            }
        )

        geo_json["features"].append(
            {
                "type": "Feature",
                "properties": {
                    "title": "Dock",
                    "Name": "Dock",
                    "description": "Charging dock location",
                    "type_name": "station",
                    "rotation": dock_rotation,
                    **DOCK_STYLE,
                    **DOCK_IMAGE,
                },
                "geometry": {"type": "Point", "coordinates": [dock_location.y, dock_location.x]},
            }
        )

    @staticmethod
    def _build_area_name_lookup(area_names: list[AreaHashNameList]) -> dict[int, str]:
        """Build a hash lookup table for area names.

        Args:
            area_names: List of AreaHashNameList objects

        Returns:
            Dictionary mapping hash to area name

        """
        return {item.hash: item.name for item in area_names}

    @staticmethod
    def _process_map_objects(
        hash_list: HashList, rtk_location: Point, area_names: dict[int, str], geo_json: GeoJSONCollection
    ) -> int:
        """Process all map objects and add them to GeoJSON.

        Args:
            hash_list: HashList object containing map data
            rtk_location: Tuple of (longitude, latitude) for rtk position
            area_names: Dictionary mapping hash to area name
            geo_json: GeoJSON collection to add features to

        Returns:
            Total number of frames processed

        """
        total_frames = 0

        # Map type names to their corresponding dictionaries in HashList
        type_mapping: dict[str, dict[int, FrameList]] = {
            "area": hash_list.area,
            "path": hash_list.path,
            "obstacle": hash_list.obstacle,
            "dump": hash_list.dump,
        }

        for type_name, map_objects in type_mapping.items():
            for hash_key, frame_list in map_objects.items():
                if not GeojsonGenerator._validate_frame_list(frame_list, hash_key, area_names):
                    continue

                local_coords = GeojsonGenerator._collect_frame_coordinates(frame_list)
                total_frames += len(frame_list.data)

                lonlat_coords = GeojsonGenerator._convert_to_lonlat_coords(local_coords, rtk_location)
                length, area = GeojsonGenerator.map_object_stats(local_coords)

                feature = GeojsonGenerator._create_feature(hash_key, frame_list, type_name, lonlat_coords, length, area)
                if feature:
                    geo_json["features"].append(feature)

        return total_frames

    @staticmethod
    def _process_mow_map_objects(hash_list: HashList, rtk_location: Point, geo_json: GeoJSONCollection) -> int:
        """Process all mow path objects and add them to GeoJSON.

        Each transaction_id in current_mow_path represents a separate mowing path
        consisting of multiple frames. A feature is generated only when all
        frames for that transaction_id have been received.
        """
        total_frames = 0

        for transaction_id, frames_by_index in hash_list.current_mow_path.items():
            if not frames_by_index:
                continue

            # Use any frame to determine total_frame and other metadata
            any_mow_path = next(iter(frames_by_index.values()))
            total_frame = any_mow_path.total_frame

            if total_frame == 0:
                continue

            # Only generate a feature when we have all frames for this transaction
            if len(frames_by_index) != total_frame:
                continue

            ordered_mow_paths = [frames_by_index[i] for i in sorted(frames_by_index.keys())]
            local_coords = GeojsonGenerator._collect_mow_frame_coordinates(ordered_mow_paths)
            total_frames += 1

            lonlat_coords = GeojsonGenerator._convert_to_lonlat_coords(local_coords, rtk_location)
            length, area = GeojsonGenerator.map_object_stats(local_coords)

            feature = GeojsonGenerator._create_mow_path_feature(any_mow_path, lonlat_coords, length, area)
            if feature:
                geo_json["features"].append(feature)

        return total_frames

    @staticmethod
    def _process_svg_map_objects(hash_list: HashList, rtk_location: Point, geo_json: GeoJSONCollection) -> None:
        """Process all SVG map objects and add them to GeoJSON."""
        for hash_key, frame_list in hash_list.svg.items():
            logger.debug(hash_key, frame_list)

    @staticmethod
    def _validate_frame_list(frame_list: FrameList, hash_key: int, area_names: dict[int, str] | None = None) -> bool:
        """Validate that frame list has complete frame data.

        Args:
            frame_list: FrameList object to validate
            hash_key: Hash key for the area
            area_names: Dictionary mapping hash to area name

        Returns:
            True if valid, False otherwise

        """
        if len(frame_list.data) != frame_list.total_frame:
            area_name = area_names.get(hash_key, "Unknown") if area_names else "Unknown"
            logger.debug(f"Error: full coord data not available for area: '{area_name}' - '{hash_key}'")
            return False
        return True

    @staticmethod
    def _collect_frame_coordinates(frame_list: FrameList) -> list[CommDataCouple]:
        """Collect coordinates from all frames in a FrameList.

        Args:
            frame_list: FrameList containing frame data

        Returns:
            List of coordinate dictionaries with 'x' and 'y' keys

        """
        local_coords: list[CommDataCouple] = []
        for frame in frame_list.data:
            if isinstance(frame, NavGetCommData):
                local_coords.extend(frame.data_couple)
            # TODO svg message needs different transform
            # elif isinstance(frame, SvgMessage):
            #     local_coords.extend(frame.)
        return local_coords

    @staticmethod
    def _collect_mow_frame_coordinates(mow_path_list: list[MowPath]) -> list[CommDataCouple]:
        """Collect coordinates from all frames in a FrameList."""
        local_coords: list[CommDataCouple] = []
        for mow_frame in mow_path_list:
            for frame in mow_frame.path_packets:
                local_coords.extend(frame.data_couple)
        return local_coords

    @staticmethod
    def _convert_to_lonlat_coords(
        local_coords: list[CommDataCouple], rtk_location: Point, x_offset: int = 0, y_offset: int = 0
    ) -> CoordinateList:
        """Convert local x,y coordinates to lon,lat coordinates.

        Args:
            local_coords: List of coordinate dictionaries with 'x' and 'y' keys
            rtk_location: Tuple of (longitude, latitude) for rtk position

        Returns:
            List of [longitude, latitude] coordinate pairs

        """
        lonlat_coords: CoordinateList = [
            list(GeojsonGenerator.lon_lat_delta(rtk_location, xy.x + x_offset, xy.y + y_offset)) for xy in local_coords
        ]
        lonlat_coords.reverse()  # GeoJSON polygons go clockwise
        return lonlat_coords

    @staticmethod
    def _create_feature(
        hash_key: int, frame_list: FrameList, type_name: str, lonlat_coords: CoordinateList, length: float, area: float
    ) -> GeoJSONFeature | None:
        """Create a GeoJSON feature from frame list data.

        Args:
            hash_key: Hash identifier for the feature
            frame_list: FrameList containing frame data
            type_name: Type name of the map object
            lonlat_coords: List of [longitude, latitude] coordinate pairs
            length: Calculated length of the feature
            area: Calculated area of the feature

        Returns:
            GeoJSON feature dictionary or None if invalid

        """
        first_frame = frame_list.data[0]
        type_id = first_frame.type
        object_name = ""
        if isinstance(first_frame, NavGetCommData):
            object_name = first_frame.name_time.name

        properties = GeojsonGenerator._create_feature_properties(
            hash_key, type_id, type_name, first_frame, length, area, object_name
        )
        geometry = GeojsonGenerator._create_feature_geometry(type_id, lonlat_coords, properties)

        if geometry is None:
            return None

        return {"type": "Feature", "properties": properties, "geometry": geometry}

    @staticmethod
    def _create_mow_path_feature(
        path_packet_list: MowPath, lonlat_coords: CoordinateList, length: float, area: float
    ) -> GeoJSONFeature | None:
        properties = GeojsonGenerator._create_feature_mow_path_properties(
            path_packet_list, length, path_packet_list.area
        )
        geometry = GeojsonGenerator._create_feature_geometry(4, lonlat_coords, properties)

        if geometry is None:
            return None

        return {"type": "Feature", "properties": properties, "geometry": geometry}

    @staticmethod
    def _create_feature_mow_path_properties(first_frame: MowPath, length: float, area: float) -> dict[str, Any]:
        """Create properties dictionary for GeoJSON feature."""
        return {
            "transaction_id": first_frame.transaction_id,
            "type_name": "mow_path",
            "total_path_num": first_frame.total_path_num,
            "length": length,
            "area": area,
            "time": first_frame.time,
        }

    @staticmethod
    def _create_feature_properties(
        hash_key: int, type_id: int, type_name: str, first_frame: Any, length: float, area: float, object_name: str = ""
    ) -> dict[str, Any]:
        """Create properties dictionary for GeoJSON feature.

        Args:
            hash_key: Hash identifier
            object_name: Name of the object
            type_id: Type ID of the feature
            type_name: Type name of the feature
            first_frame: First frame from the FrameList
            length: Calculated length
            area: Calculated area

        Returns:
            Properties dictionary

        """
        return {
            "hash": hash_key,
            "title": object_name,
            "Name": object_name,
            "description": "description <b>test</b>",
            "type_id": type_id,
            "type_name": type_name,
            "parent_hash_a": first_frame.paternal_hash_a,
            "parent_hash_b": first_frame.paternal_hash_b,
            "length": length,
            "area": area,
        }

    @staticmethod
    def _create_feature_geometry(
        type_id: int, lonlat_coords: CoordinateList, properties: dict[str, Any]
    ) -> dict[str, Any] | None:
        """Create geometry dictionary for GeoJSON feature based on type.

        Args:
            type_id: Type ID determining geometry type
            lonlat_coords: List of [longitude, latitude] coordinate pairs
            properties: Properties dictionary to modify with style information

        Returns:
            Geometry dictionary or None if invalid type

        """
        if type_id == TYPE_MOWING_ZONE:
            properties.update(AREA_STYLE)
            return {"type": "Polygon", "coordinates": [lonlat_coords]}
        elif type_id == TYPE_OBSTACLE:
            properties.update(OBSTACLE_STYLE)
            return {"type": "Polygon", "coordinates": [lonlat_coords]}
        elif type_id == TYPE_PATH and len(lonlat_coords) > 1:
            properties.update(PATH_STYLE)
            return {"type": "LineString", "coordinates": lonlat_coords}
        elif type_id == TYPE_MOW_PATH and len(lonlat_coords) > 1:
            properties["color"] = "green"
            return {"type": "LineString", "coordinates": lonlat_coords}
        else:
            return None  # Point (ignore)

    @staticmethod
    def _save_geojson(geoJSON: GeoJSONCollection) -> None:
        """Save GeoJSON data to file.

        Args:
            geoJSON: GeoJSON collection to save

        """
        with open("areas.json", "w") as json_file:
            json.dump(geoJSON, json_file, indent=2)

    @staticmethod
    def lon_lat_delta(rtk: Point, x: float, y: float) -> Coordinate:
        """Add delta (in meters) to lon/lat, return new lon/lat.

        Args:
            lon: Longitude in degrees
            lat: Latitude in degrees
            x: X offset in meters
            y: Y offset in meters

        Returns:
            Tuple of (new_longitude, new_latitude)

        """
        new_lon = rtk.y + (x / (METERS_PER_DEGREE * math.cos(math.radians(rtk.x))))
        new_lat = rtk.x + (y / METERS_PER_DEGREE)
        return new_lon, new_lat

    @staticmethod
    def map_object_stats(coords: list[CommDataCouple]) -> Coordinate:
        """Calculate length and area statistics for map object coordinates.

        Args:
            coords: List of coordinate dictionaries with 'x' and 'y' keys

        Returns:
            Tuple of (length, area) in meters and square meters

        """
        # Point Object
        if len(coords) < 2:
            return 0.0, 0.0

        def distance(p1: CommDataCouple, p2: CommDataCouple) -> float:
            """Calculate Euclidean distance between two points."""
            return math.sqrt((p2.x - p1.x) ** 2 + (p2.y - p1.y) ** 2)

        length = sum(distance(coords[i], coords[i + 1]) for i in range(len(coords) - 1))

        # Open line
        if coords[0] != coords[-1]:
            return length, 0.0

        # Closed Polygon - Calculate area using shoelace formula
        area = 0.5 * abs(
            sum(coords[i].x * coords[i + 1].y - coords[i + 1].x * coords[i].y for i in range(len(coords) - 1))
        )

        return length, area

    @staticmethod
    def is_point_in_polygon(x: float, y: float, poly: list[list[float]]) -> bool:
        """Test if a point is inside a polygon using ray casting algorithm.

        Args:
            x: X coordinate of the point
            y: Y coordinate of the point
            poly: Polygon as list of [x, y] coordinate pairs

        Returns:
            True if point is inside polygon, False otherwise

        """
        return (
            sum(
                (y > poly[i][1]) != (y > poly[(i + 1) % len(poly)][1])
                and (
                    x
                    < (poly[(i + 1) % len(poly)][0] - poly[i][0])
                    * (y - poly[i][1])
                    / (poly[(i + 1) % len(poly)][1] - poly[i][1])
                    + poly[i][0]
                )
                for i in range(len(poly))
            )
            % 2
            == 1
        )



from typing import TypeVar, Union, get_args, get_origin

from pymammotion.http.model.http import Response

T = TypeVar("T")


def deserialize_data(value, target_type):
    """Deserialize data into a specified target type.

    The function handles deserialization of basic types, lists, and unions. It
    recursively processes list elements and supports optional types by handling
    Union[T, None]. For custom types with a `from_dict` method, it calls this
    method for deserialization. If the target type is unknown or unsupported, it
    returns the value unchanged.

    Args:
        value: The data to be deserialized.
        target_type (type): The desired type into which the data should be deserialized.

    Returns:
        The deserialized data in the specified target type.

    """
    if value is None:
        return None

    origin = get_origin(target_type)
    args = get_args(target_type)

    if origin is list and args:
        item_type = args[0]
        return [deserialize_data(v, item_type) for v in value]

    if origin is Union:
        # Support Optional[T] = Union[T, None]
        non_none_types = [t for t in args if t is not type(None)]
        if len(non_none_types) == 1:
            target = non_none_types[0]
            # Handle Response[list[type]] case
            if get_origin(target) is list and get_args(target):
                item_type = get_args(target)[0]
                return [deserialize_data(v, item_type) for v in value]
            return deserialize_data(value, target)

    if hasattr(target_type, "from_dict"):
        return target_type.from_dict(value)

    return value  # fallback: unknown type, leave as-is


def response_factory(response_cls: type[Response[T]], raw_dict: dict) -> Response[T]:
    # Extract the type of the generic `data` field
    """Create a Response instance from a dictionary."""
    data_type = get_args(response_cls)[0] if get_args(response_cls) else None

    if data_type:
        data_value = deserialize_data(raw_dict.get("data"), data_type)
        return Response(code=raw_dict["code"], msg=raw_dict["msg"], data=data_value)
    else:
        return response_cls.from_dict(raw_dict)



# === sendOrderMsg_Nav ===
from abc import ABC
import logging
import time

from pymammotion.data.model import GenerateRouteInformation
from pymammotion.data.model.hash_list import Plan, SvgMessage
from pymammotion.data.model.region_data import RegionData
from pymammotion.mammotion.commands.abstract_message import AbstractMessage
from pymammotion.proto import (
    AppRequestCoverPathsT,
    LubaMsg,
    MctlNav,
    MsgAttr,
    MsgCmdType,
    MsgDevice,
    NavGetCommData,
    NavGetHashList,
    NavMapNameMsg,
    NavPlanJobSet,
    NavPlanTaskExecute,
    NavReqCoverPath,
    NavSysParamMsg,
    NavTaskCtrl,
    NavUnableTimeSet,
    NavUploadZigZagResultAck,
    SimulationCmdData,
    SvgMessageAckT,
    SvgMessageT,
    WorkReportCmdData,
    WorkReportUpdateCmd,
)

logger = logging.getLogger(__name__)


class MessageNavigation(AbstractMessage, ABC):
    def send_order_msg_nav(self, build: MctlNav) -> bytes:
        luba_msg = LubaMsg(
            msgtype=MsgCmdType.NAV,
            sender=MsgDevice.DEV_MOBILEAPP,
            rcver=self.get_msg_device(MsgCmdType.NAV, MsgDevice.DEV_MAINCTL),
            msgattr=MsgAttr.REQ,
            seqs=self.seqs.increment_and_get() & 255,
            version=1,
            subtype=self.user_account,
            nav=build,
            timestamp=round(time.time() * 1000),
        )

        return luba_msg.SerializeToString()

    def allpowerfull_rw_adapter_x3(self, rw_id: int, context: int, rw: int) -> bytes:
        build = MctlNav(nav_sys_param_cmd=NavSysParamMsg(id=rw_id, context=context, rw=rw))
        logger.debug(f"Send command--x3 general read and write command id={rw_id}, context={context}, rw={rw}")
        return self.send_order_msg_nav(build)

    def along_border(self) -> bytes:
        build = MctlNav(todev_edgecmd=1)
        logger.debug("Send command--along the edge command")
        return self.send_order_msg_nav(build)

    def start_draw_border(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=0, type=0))
        logger.debug("Send command--Start drawing boundary command")
        return self.send_order_msg_nav(build)

    def enter_dumping_status(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=14, type=12))
        logger.debug("Send command--Enter grass collection status")
        return self.send_order_msg_nav(build)

    def add_dump_point(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=0, type=12))
        logger.debug("Send command--Add grass collection point")
        return self.send_order_msg_nav(build)

    def revoke_dump_point(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=6, type=12))
        logger.debug("Send command--Revoke grass collection point")
        return self.send_order_msg_nav(build)

    def exit_dumping_status(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=1, type=12))
        logger.debug("Send command--Exit grass collection setting status")
        return self.send_order_msg_nav(build)

    def out_drop_dumping_add(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=15, type=12))
        logger.debug("Send command--Complete external grass collection point marking operation")
        return self.send_order_msg_nav(build)

    def recover_dumping(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=12, type=12))
        logger.debug("Send command--Recover grass collection operation")
        return self.send_order_msg_nav(build)

    def start_draw_barrier(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=0, type=1))
        logger.debug("Sending command - Draw obstacle command")
        return self.send_order_msg_nav(build)

    def start_erase(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=4, type=0))
        logger.debug("Sending command - Start erase command - Bluetooth")
        return self.send_order_msg_nav(build)

    def end_erase(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=5, type=0))
        logger.debug("Sending command - End erase command")
        return self.send_order_msg_nav(build)

    def cancel_erase(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=7, type=0))
        logger.debug("Sending command - Cancel erase command")
        return self.send_order_msg_nav(build)

    def start_channel_line(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=0, type=2))
        logger.debug("Sending command - Start drawing channel line command")
        return self.send_order_msg_nav(build)

    def save_task(self) -> bytes:
        build = MctlNav(todev_save_task=1)
        logger.debug("Sending command - Save task command")
        return self.send_order_msg_nav(build)

    def set_edit_boundary(self, action: int) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=action, type=0))
        logger.debug(f"Sending secondary editing command action={action}")
        return self.send_order_msg_nav(build)

    def set_data_synchronization(self, type: int) -> bytes:
        logger.debug(f"Sync data ==================== Sending ============ Restore command: {type}")
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=12, type=type))
        logger.debug("Sync data ==================== Sending ============ Restore command")
        return self.send_order_msg_nav(build)

    def send_plan(self, plan_bean: Plan) -> bytes:
        build = MctlNav(
            todev_planjob_set=NavPlanJobSet(
                pver=plan_bean.pver,
                sub_cmd=plan_bean.sub_cmd,
                area=plan_bean.area,
                work_time=plan_bean.work_time,
                version=plan_bean.version,
                id=plan_bean.id,
                user_id=plan_bean.user_id,
                device_id=plan_bean.device_id,
                plan_id=plan_bean.plan_id,
                task_id=plan_bean.task_id,
                job_id=plan_bean.job_id,
                start_time=plan_bean.start_time,
                end_time=plan_bean.end_time,
                week=plan_bean.week,
                knife_height=plan_bean.knife_height,
                model=plan_bean.model,
                edge_mode=plan_bean.edge_mode,
                required_time=plan_bean.required_time,
                route_angle=plan_bean.route_angle,
                route_model=plan_bean.route_model,
                route_spacing=plan_bean.route_spacing,
                ultrasonic_barrier=plan_bean.ultrasonic_barrier,
                total_plan_num=plan_bean.total_plan_num,
                plan_index=plan_bean.plan_index,
                result=plan_bean.result,
                speed=plan_bean.speed,
                task_name=plan_bean.task_name,
                job_name=plan_bean.job_name,
                zone_hashs=plan_bean.zone_hashs,
                reserved=plan_bean.reserved,
            )
        )
        logger.debug(f"Send read job plan command planBean={plan_bean}")
        return self.send_order_msg_nav(build)

    def send_schedule(self, plan_bean: Plan) -> bytes:
        build = NavPlanJobSet(
            pver=plan_bean.pver,
            sub_cmd=plan_bean.sub_cmd,
            area=plan_bean.area,
            device_id=plan_bean.device_id,
            work_time=plan_bean.work_time,
            version=plan_bean.version,
            id=plan_bean.id,
            user_id=plan_bean.user_id,
            plan_id=plan_bean.plan_id,
            task_id=plan_bean.task_id,
            job_id=plan_bean.job_id,
            start_time=plan_bean.start_time,
            end_time=plan_bean.end_time,
            week=plan_bean.week,
            knife_height=plan_bean.knife_height,
            model=plan_bean.model,
            edge_mode=plan_bean.edge_mode,
            required_time=plan_bean.required_time,
            route_angle=plan_bean.route_angle,
            route_model=plan_bean.route_model,
            route_spacing=plan_bean.route_spacing,
            ultrasonic_barrier=plan_bean.ultrasonic_barrier,
            total_plan_num=plan_bean.total_plan_num,
            plan_index=plan_bean.plan_index,
            result=plan_bean.result,
            speed=plan_bean.speed,
            task_name=plan_bean.task_name,
            job_name=plan_bean.job_name,
            zone_hashs=plan_bean.zone_hashs,
            reserved=plan_bean.reserved,
            weeks=plan_bean.weeks,
            start_date=plan_bean.start_date,
            trigger_type=plan_bean.trigger_type,
            day=plan_bean.day,
            toward_included_angle=plan_bean.toward_included_angle,
            toward_mode=0,
        )
        logger.debug(f"Send read job plan command planBean={plan_bean}")
        return self.send_order_msg_nav(MctlNav(todev_planjob_set=build))

    def single_schedule(self, plan_id: str) -> bytes:
        return self.send_order_msg_nav(MctlNav(plan_task_execute=NavPlanTaskExecute(sub_cmd=1, id=plan_id)))

    def read_plan(self, sub_cmd: int, plan_index: int = 0) -> bytes:
        build = MctlNav(todev_planjob_set=NavPlanJobSet(sub_cmd=sub_cmd, plan_index=plan_index))
        logger.debug(f"Send read job plan command cmd={sub_cmd} PlanIndex = {plan_index}")
        return self.send_order_msg_nav(build)

    def delete_plan(self, sub_cmd: int, plan_id: str) -> bytes:
        build = MctlNav(todev_planjob_set=NavPlanJobSet(sub_cmd=sub_cmd, plan_id=plan_id))
        logger.debug(f"Send command--Send delete job plan command cmd={sub_cmd} planId = {plan_id}")
        return self.send_order_msg_nav(build)

    def set_plan_unable_time(self, sub_cmd: int, device_id: str, unable_end_time: str, unable_start_time: str) -> bytes:
        build = NavUnableTimeSet(
            sub_cmd=sub_cmd,
            device_id=device_id,
            unable_end_time=unable_end_time,
            result=0,
            reserved="0",
            unable_start_time=unable_start_time,
        )
        logger.debug(f"{self.get_device_name()} Set forbidden time===={build}")
        return self.send_order_msg_nav(MctlNav(todev_unable_time_set=build))

    def read_plan_unable_time(self, sub_cmd: int) -> bytes:
        build = NavUnableTimeSet(sub_cmd=sub_cmd)
        build2 = MctlNav(todev_unable_time_set=build)
        logger.debug(f"Send command--Read plan time {sub_cmd}")
        return self.send_order_msg_nav(build2)

    def read_job_not_not_disturb(self) -> bytes:
        build = NavUnableTimeSet(sub_cmd=2)
        build2 = MctlNav(todev_unable_time_set=build)
        logger.debug(f"Send command--Read job dnd {2}")
        return self.send_order_msg_nav(build2)

    def job_animal_protect_read(self) -> bytes:
        """Read animal protection settings."""
        build = NavUnableTimeSet(sub_cmd=2, trigger=99)
        build2 = MctlNav(todev_unable_time_set=build)
        logger.debug(f"Send command - Read job do not disturb time subCmd2 {build}")
        return self.send_order_msg_nav(build2)

    def job_do_not_disturb(self, unable_start_time: str, unable_end_time: str) -> bytes:
        """Set do not disturb time period."""
        build = MctlNav(
            todev_unable_time_set=NavUnableTimeSet(
                sub_cmd=1, trigger=1, unable_start_time=unable_start_time, unable_end_time=unable_end_time
            )
        )
        logger.debug(f"Send command - Set job do not disturb time: {unable_start_time} - {unable_end_time}")
        return self.send_order_msg_nav(build)

    def job_do_not_disturb_del(self) -> bytes:
        """Delete do not disturb settings."""
        build = MctlNav(todev_unable_time_set=NavUnableTimeSet(sub_cmd=1, trigger=0))
        logger.debug("Send command - Turn off do not disturb time")
        return self.send_order_msg_nav(build)

    def query_job_history(self) -> bytes:
        return self.send_order_msg_nav(MctlNav(todev_work_report_update_cmd=WorkReportUpdateCmd(sub_cmd=1)))

    def request_job_history(self, num: int) -> bytes:
        return self.send_order_msg_nav(MctlNav(todev_work_report_cmd=WorkReportCmdData(sub_cmd=1, get_info_num=num)))

    def leave_dock(self) -> bytes:
        build = MctlNav(todev_one_touch_leave_pile=1)
        logger.debug("Send command--One-click automatic undocking")
        return self.send_order_msg_nav(build)

    def get_area_name_list(self, device_id: str) -> bytes:
        # Build the NavMapNameMsg with the specified parameters
        mctl_nav = MctlNav(
            toapp_map_name_msg=NavMapNameMsg(
                hash=0,
                result=0,
                device_id=device_id,  # iot_id
                rw=0,
            )
        )

        # Send the message with the specified ID and acknowledge flag
        logger.debug("Send command--Get area name list")
        return self.send_order_msg_nav(mctl_nav)

    def set_area_name(self, device_id: str, hash_id: int, name: str) -> bytes:
        # Build the NavMapNameMsg with the specified parameters
        mctl_nav = MctlNav(
            toapp_map_name_msg=NavMapNameMsg(hash=hash_id, name=name, result=0, device_id=device_id, rw=1)
        )

        # Send the message with the specified ID and acknowledge flag
        logger.debug("Send command--Get area name list")
        return self.send_order_msg_nav(mctl_nav)

    def get_all_boundary_hash_list(self, sub_cmd: int) -> bytes:
        build = MctlNav(todev_gethash=NavGetHashList(pver=1, sub_cmd=sub_cmd))
        logger.debug(f"Area loading=====================:Get area hash list:{sub_cmd}")
        return self.send_order_msg_nav(build)

    def get_hash_response(self, total_frame: int, current_frame: int) -> bytes:
        build = MctlNav(
            todev_gethash=NavGetHashList(pver=1, sub_cmd=2, current_frame=current_frame, total_frame=total_frame)
        )
        logger.debug(
            f"Send command--208 Response hash list command totalFrame={total_frame},currentFrame={current_frame}"
        )
        return self.send_order_msg_nav(build)

    def synchronize_hash_data(self, hash_num: int) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=8, hash=hash_num, sub_cmd=1))
        logger.debug(f"Send command--209,hash synchronize area data hash:{hash_num}")
        return self.send_order_msg_nav(build)

    def get_area_to_be_transferred(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=8, sub_cmd=1, type=3))
        logger.debug("Send command--Get transfer area before charging pile")
        return self.send_order_msg_nav(build)

    def get_regional_data(self, regional_data: RegionData) -> bytes:
        build = MctlNav(
            todev_get_commondata=NavGetCommData(
                pver=1,
                action=regional_data.action,
                type=regional_data.type,
                hash=regional_data.hash,
                total_frame=regional_data.total_frame,
                current_frame=regional_data.current_frame,
                sub_cmd=2,
            )
        )
        logger.debug("Area loading=====================:Response area data")
        return self.send_order_msg_nav(build)

    def indoor_simulation(self, flag: int) -> bytes:
        build = MctlNav(simulation_cmd=SimulationCmdData(sub_cmd=flag))
        logger.debug(f"Send command--Send indoor simulation command flag={flag}")
        return self.send_order_msg_nav(build)

    def send_tools_order(self, param_id: int, values: list[int]) -> bytes:
        build = MctlNav(simulation_cmd=SimulationCmdData(sub_cmd=2, param_id=param_id, param_value=values))
        logger.debug(f"Send command--Send tool command id={param_id},values={values}")
        return self.send_order_msg_nav(build)

    def end_draw_border(self, type: int) -> bytes | None:
        if type == -1:
            return None
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=1, type=type))
        logger.debug(f"Send command--End drawing boundary, obstacle, channel command type={type}")
        return self.send_order_msg_nav(build)

    def cancel_current_record(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=7, sub_cmd=0))
        logger.debug("Send command--Cancel current recording (boundary, obstacle)")
        return self.send_order_msg_nav(build)

    def delete_map_elements(self, type: int, hash_num: int) -> bytes | None:
        if type == -1:
            return None
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=6, type=type, hash=hash_num))
        logger.debug(f"Send command--Delete boundary or obstacle or channel command type={type},hash={hash}")
        return self.send_order_msg_nav(build)

    def delete_charge_point(self) -> bytes:
        logger.debug("Delete charging pile")
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=6, type=5))
        logger.debug("Send command--Delete charging pile location and reset")
        return self.send_order_msg_nav(build)

    def confirm_base_station(self) -> bytes:
        logger.debug("Reset base station")
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=2, type=7))
        logger.debug("Send command--Confirm no modification to base station")
        return self.send_order_msg_nav(build)

    def delete_all(self) -> bytes:
        build = MctlNav(todev_get_commondata=NavGetCommData(pver=1, action=6, type=6))
        logger.debug("Send command--Clear job data")
        return self.send_order_msg_nav(build)

    def generate_route_information(self, generate_route_information: GenerateRouteInformation) -> bytes:
        logger.debug(f"Generate route data source:{generate_route_information}")

        build = NavReqCoverPath(
            pver=1,
            sub_cmd=0,
            zone_hashs=generate_route_information.one_hashs,
            job_mode=int(generate_route_information.job_mode),
            edge_mode=int(generate_route_information.edge_mode),
            knife_height=int(generate_route_information.blade_height),
            speed=float(generate_route_information.speed),
            ultra_wave=int(generate_route_information.ultra_wave),
            channel_width=int(generate_route_information.channel_width),
            channel_mode=int(generate_route_information.channel_mode),
            toward=int(generate_route_information.toward),
            toward_included_angle=int(generate_route_information.toward_included_angle),  # luba 2 yuka only
            toward_mode=int(generate_route_information.toward_mode),  # luba 2 yuka only
            reserved=generate_route_information.path_order,
        )
        logger.debug(f"{self.get_device_name()}Generate route====={build}")
        logger.debug(f"Send command--Generate route information generateRouteInformation={
        generate_route_information}")
        return self.send_order_msg_nav(MctlNav(bidire_reqconver_path=build))

    def modify_route_information(self, generate_route_information: GenerateRouteInformation) -> bytes:
        logger.debug(f"Generate route data source: {generate_route_information}")
        build = NavReqCoverPath(
            pver=1,
            sub_cmd=3,
            zone_hashs=generate_route_information.one_hashs,
            job_mode=int(generate_route_information.job_mode),
            edge_mode=int(generate_route_information.edge_mode),
            knife_height=int(generate_route_information.blade_height),
            speed=float(generate_route_information.speed),
            ultra_wave=int(generate_route_information.ultra_wave),
            channel_width=int(generate_route_information.channel_width),
            channel_mode=int(generate_route_information.channel_mode),
            toward=int(generate_route_information.toward),
            reserved=generate_route_information.path_order,
        )
        logger.debug(f"{self.get_device_name()} Generate route ===== {build}")
        logger.debug(f"Send command -- Modify route parameters generate_route_information={
        generate_route_information}")
        return self.send_order_msg_nav(MctlNav(bidire_reqconver_path=build))

    def end_generate_route_information(self) -> bytes:
        build = NavReqCoverPath(pver=1, sub_cmd=9)
        logger.debug(f"{self.get_device_name()} Generate route ===== {build}")
        build2 = MctlNav(bidire_reqconver_path=build)
        logger.debug("Send command -- End generating route information generate_route_information=")
        return self.send_order_msg_nav(build2)

    def query_generate_route_information(self) -> bytes:
        build = NavReqCoverPath(pver=1, sub_cmd=2)
        logger.debug(f"{self.get_device_name(
        )} Send command -- Get route configuration information generate_route_information={build}")
        build2 = MctlNav(bidire_reqconver_path=build)
        return self.send_order_msg_nav(build2)

    def get_line_info(self, current_hash: int) -> bytes:
        logger.debug(f"Sending==========Get route command: {current_hash}")
        build = MctlNav(todev_zigzag_ack=NavUploadZigZagResultAck(pver=1, current_hash=current_hash, sub_cmd=0))
        logger.debug(f"Sending command--Get route data corresponding to hash={current_hash}")
        return self.send_order_msg_nav(build)

    def get_line_info_list(self, hash_list: list[int], transaction_id: int) -> bytes:
        """Get route information (mow path) corresponding to the specified hash list based on time.
        e.g transaction_id = int(time.time() * 1000)
        """
        logger.debug(f"Sending==========Get route command: {hash_list}")

        build = MctlNav(
            app_request_cover_paths=AppRequestCoverPathsT(
                pver=1, hash_list=hash_list, transaction_id=transaction_id, sub_cmd=0
            )
        )
        logger.debug(f"Sending command--Get route data corresponding to hash={hash_list}")
        return self.send_order_msg_nav(build)

    def start_job(self) -> bytes:
        logger.debug("Sending==========Start job command")
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=1, result=0))
        logger.debug("Sending command--Start job")
        return self.send_order_msg_nav(build)

    def cancel_return_to_dock(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=12, result=0))
        logger.debug("Send command - Cancel return to charge")
        return self.send_order_msg_nav(build)

    def cancel_job(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=4, result=0))
        logger.debug("Send command - End job")
        return self.send_order_msg_nav(build)

    def return_to_dock(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=5, result=0))
        logger.debug("Send command - Return to charge command")
        return self.send_order_msg_nav(build)

    def pause_execute_task(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=2, result=0))
        logger.debug("Send command - Pause command")
        return self.send_order_msg_nav(build)

    def re_charge_test(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=10, result=0))
        logger.debug("Send command - Return to charge test command")
        return self.send_order_msg_nav(build)

    def fast_aotu_test(self, action: int) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=action, result=0))
        logger.debug("Send command - One-click automation test")
        return self.send_order_msg_nav(build)

    def resume_execute_task(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=3, result=0))
        logger.debug("Send command - Cancel pause command")
        return self.send_order_msg_nav(build)

    def break_point_continue(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=7, result=0))
        logger.debug("Send command - Continue from breakpoint")
        return self.send_order_msg_nav(build)

    def break_point_anywhere_continue(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=1, action=9, result=0))
        logger.debug("Send command - Continue from current vehicle position")
        return self.send_order_msg_nav(build)

    def reset_base_station(self) -> bytes:
        build = MctlNav(todev_taskctrl=NavTaskCtrl(type=3, action=1, result=0))
        logger.debug("Send command - Reset charging pile, base station position")
        return self.send_order_msg_nav(build)

    def send_svg_data(self, svg_message: SvgMessage) -> bytes:
        """Send SVG data to the device."""
        build = MctlNav(
            todev_svg_msg=SvgMessageAckT(
                pver=1,
                sub_cmd=svg_message.sub_cmd,
                total_frame=svg_message.total_frame,
                current_frame=svg_message.current_frame,
                data_hash=svg_message.data_hash,
                paternal_hash_a=svg_message.paternal_hash_a,
                result=svg_message.result,
                svg_message=SvgMessageT(
                    hide_svg=svg_message.svg_message.hide_svg,
                    svg_file_data=svg_message.svg_message.svg_file_data,
                    svg_file_name=svg_message.svg_message.svg_file_name,
                    data_count=svg_message.svg_message.data_count,
                    name_count=svg_message.svg_message.name_count,
                    base_height_m=svg_message.svg_message.base_height_m,
                    base_width_m=svg_message.svg_message.base_width_m,
                    base_width_pix=svg_message.svg_message.base_width_pix,
                    base_height_pix=svg_message.svg_message.base_height_pix,
                    x_move=svg_message.svg_message.x_move,
                    y_move=svg_message.svg_message.y_move,
                    scale=svg_message.svg_message.scale,
                    rotate=svg_message.svg_message.rotate,
                ),
            )
        )

        return self.send_order_msg_nav(build)



"""RTK device class without map synchronization callbacks."""

import asyncio
import logging
from typing import Any

from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model.device import RTKDevice
from pymammotion.data.model.raw_data import RawMowerData

_LOGGER = logging.getLogger(__name__)


class MammotionRTKDevice:
    """RTK device without map synchronization - simpler than mowers."""

    def __init__(self, cloud_device: Device, rtk_state: RTKDevice) -> None:
        """Initialize MammotionRTKDevice."""
        self.loop = asyncio.get_event_loop()
        self._rtk_device = rtk_state
        self._raw_data = dict()
        self._raw_mower_data: RawMowerData = RawMowerData()
        self._notify_future: asyncio.Future[bytes] | None = None
        self._cloud_device = cloud_device

    @property
    def rtk(self) -> RTKDevice:
        """Get the RTK device state."""
        return self._rtk_device

    @property
    def raw_data(self) -> dict[str, Any]:
        """Get the raw data of the device."""
        return self._raw_data

    async def command(self, key: str, **kwargs: Any) -> bytes | None:
        """Send a command to the device."""
        return await self.queue_command(key, **kwargs)

    async def queue_command(self, key: str, **kwargs: Any) -> bytes | None:
        """Queue commands to RTK device - to be implemented by connection-specific subclasses."""
        raise NotImplementedError("Subclasses must implement queue_command")

    async def _ble_sync(self) -> None:
        """Send ble sync command - to be implemented by connection-specific subclasses."""
        raise NotImplementedError("Subclasses must implement _ble_sync")

    def stop(self) -> None:
        """Stop everything ready for destroying - to be implemented by connection-specific subclasses."""
        raise NotImplementedError("Subclasses must implement stop")



"""Mammotion devices module."""

from .mammotion import Mammotion, MammotionDeviceManager
from .mammotion_bluetooth import MammotionBaseBLEDevice
from .mammotion_cloud import MammotionBaseCloudDevice, MammotionCloud
from .mammotion_mower_ble import MammotionMowerBLEDevice
from .mammotion_mower_cloud import MammotionMowerCloudDevice
from .mower_device import MammotionMowerDevice
from .mower_manager import MammotionMowerDeviceManager
from .rtk_ble import MammotionRTKBLEDevice
from .rtk_cloud import MammotionRTKCloudDevice
from .rtk_device import MammotionRTKDevice
from .rtk_manager import MammotionRTKDeviceManager

__all__ = [
    "Mammotion",
    "MammotionDeviceManager",
    "MammotionMowerDeviceManager",
    "MammotionBaseBLEDevice",
    "MammotionBaseCloudDevice",
    "MammotionCloud",
    "MammotionMowerBLEDevice",
    "MammotionMowerCloudDevice",
    "MammotionMowerDevice",
    "MammotionRTKBLEDevice",
    "MammotionRTKCloudDevice",
    "MammotionRTKDevice",
    "MammotionRTKDeviceManager",
]



"""Module for interacting with Aliyun Cloud IoT Gateway."""

import asyncio
import base64
import hashlib
import hmac
import itertools
import json
from json.decoder import JSONDecodeError
from logging import getLogger
import random
import string
import time
import uuid

from aiohttp import ClientSession, ConnectionTimeoutError
from alibabacloud_iot_api_gateway.models import CommonParams, Config, IoTApiRequest
from alibabacloud_tea_util.client import Client as UtilClient
from alibabacloud_tea_util.models import RuntimeOptions
from Tea.exceptions import UnretryableException

from pymammotion.aliyun.client import Client
from pymammotion.aliyun.model.aep_response import AepResponse
from pymammotion.aliyun.model.connect_response import ConnectResponse
from pymammotion.aliyun.model.dev_by_account_response import ListingDevAccountResponse
from pymammotion.aliyun.model.login_by_oauth_response import LoginByOAuthResponse
from pymammotion.aliyun.model.regions_response import RegionResponse
from pymammotion.aliyun.model.session_by_authcode_response import SessionByAuthCodeResponse
from pymammotion.aliyun.model.thing_response import ThingPropertiesResponse
from pymammotion.aliyun.regions import region_mappings
from pymammotion.const import ALIYUN_DOMAIN, APP_KEY, APP_SECRET, APP_VERSION
from pymammotion.http.http import MammotionHTTP
from pymammotion.utility.datatype_converter import DatatypeConverter

logger = getLogger(__name__)

MOVE_HEADERS = (
    "x-ca-signature",
    "x-ca-signature-headers",
    "accept",
    "content-md5",
    "content-type",
    "date",
    "host",
    "token",
    "user-agent",
)


class SetupException(Exception):
    """Raise when mqtt expires token or token is invalid."""

    def __init__(self, *args: object) -> None:
        super().__init__(args)
        self.iot_id = args[1]


class AuthRefreshException(Exception):
    """Raise exception when library cannot refresh token."""


class DeviceOfflineException(Exception):
    """Raise exception when device is offline."""

    def __init__(self, *args: object) -> None:
        super().__init__(args)
        self.iot_id = args[1]


class FailedRequestException(Exception):
    """Raise exception when request response is bad."""

    def __init__(self, *args: object) -> None:
        super().__init__(args)
        self.iot_id = args[0]


class NoConnectionException(UnretryableException):
    """Raise exception when device is unreachable."""


class GatewayTimeoutException(Exception):
    """Raise exception when the gateway times out."""

    def __init__(self, *args: object) -> None:
        super().__init__(args)
        self.iot_id = args[1]


class TooManyRequestsException(Exception):
    """Raise exception when the gateway times out."""

    def __init__(self, *args: object) -> None:
        super().__init__(args)
        self.iot_id = args[1]


class LoginException(Exception):
    """Raise exception when library cannot log in."""


class CheckSessionException(Exception):
    """Raise exception when checking session results in a failure."""


EXPIRED_CREDENTIAL_EXCEPTIONS = (CheckSessionException, SetupException)


class CloudIOTGateway:
    """Class for interacting with Aliyun Cloud IoT Gateway."""

    _client_id = ""
    _device_sn = ""
    _utdid = ""

    converter = DatatypeConverter()

    def __init__(
        self,
        mammotion_http: MammotionHTTP,
        connect_response: ConnectResponse | None = None,
        login_by_oauth_response: LoginByOAuthResponse | None = None,
        aep_response: AepResponse | None = None,
        session_by_authcode_response: SessionByAuthCodeResponse | None = None,
        region_response: RegionResponse | None = None,
        dev_by_account: ListingDevAccountResponse | None = None,
    ) -> None:
        """Initialize the CloudIOTGateway."""
        self.mammotion_http: MammotionHTTP = mammotion_http
        self._app_key = APP_KEY
        self._app_secret = APP_SECRET
        self.domain = ALIYUN_DOMAIN
        self.message_delay = 1
        self._client_id = self.generate_hardware_string(8)  # 8 characters
        self._device_sn = self.generate_hardware_string(32)  # 32 characters
        self._utdid = self.generate_hardware_string(32)  # 32 characters
        self._connect_response = connect_response
        self._login_by_oauth_response = login_by_oauth_response
        self._aep_response = aep_response
        self._session_by_authcode_response = session_by_authcode_response
        self._region_response = region_response
        self._devices_by_account_response = dev_by_account
        self._iot_token_issued_at = int(time.time())
        if self._session_by_authcode_response:
            self._iot_token_issued_at = (
                self._session_by_authcode_response.token_issued_at
                if self._session_by_authcode_response.token_issued_at is not None
                else int(time.time())
            )

    @staticmethod
    def generate_random_string(length: int) -> str:
        """Generate a random string of specified length."""
        characters = string.ascii_letters + string.digits
        return "".join(random.choice(characters) for _ in range(length))

    @staticmethod
    def generate_hardware_string(length: int) -> str:
        """Generate hardware string that is consistent per device."""
        hashed_uuid = hashlib.sha1(f"{uuid.getnode()}".encode()).hexdigest()
        return "".join(itertools.islice(itertools.cycle(hashed_uuid), length))

    @staticmethod
    def parse_json_response(response_body_str: str) -> dict:
        try:
            return json.loads(response_body_str) if response_body_str is not None else {}
        except JSONDecodeError:
            logger.error("Couldn't decode message %s", response_body_str)
            return {"code": 22000}

    def sign(self, data: dict) -> str:
        """Generate signature for the given data."""
        keys = ["appKey", "clientId", "deviceSn", "timestamp"]
        concatenated_str = ""
        for key in keys:
            concatenated_str += f"{key}{data.get(key, '')}"

        logger.debug("sign(), toSignStr = %s", concatenated_str)

        return hmac.new(
            self._app_secret.encode("utf-8"),
            concatenated_str.encode("utf-8"),
            hashlib.sha1,
        ).hexdigest()

    async def get_region(self, country_code: str) -> RegionResponse:
        """Get the region based on country code and auth code."""
        auth_code = self.mammotion_http.login_info.authorization_code

        if self._region_response is not None:
            return self._region_response

        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self.domain,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.2", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "authCode": auth_code,
                "type": "THIRD_AUTHCODE",
                "countryCode": country_code,
            },
            request=request,
            version="1.0",
        )

        # send request
        try:
            response = await client.async_do_request(
                "/living/account/region/get", "https", "POST", None, body, RuntimeOptions()
            )
            logger.debug(response.status_message)
            logger.debug(response.headers)
            logger.debug(response.status_code)
            logger.debug(response.body)
        except ConnectionTimeoutError:
            body = {"data": {}, "code": 200}

            region = region_mappings.get(country_code, "US")
            body["data"]["shortRegionId"] = region
            body["data"]["regionEnglishName"] = ""
            body["data"]["oaApiGatewayEndpoint"] = f"living-account.{region}.aliyuncs.com"
            body["data"]["regionId"] = region
            body["data"]["mqttEndpoint"] = f"public.itls.{region}.aliyuncs.com:1883"
            body["data"]["pushChannelEndpoint"] = f"living-accs.{region}.aliyuncs.com"
            body["data"]["apiGatewayEndpoint"] = f"{region}.api-iot.aliyuncs.com"

            RegionResponse.from_dict(body)
            return body
        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in getting regions: " + response_body_dict["msg"])

        self._region_response = RegionResponse.from_dict(response_body_dict)
        logger.debug("Endpoint: %s", self._region_response.data.mqttEndpoint)

        return response.body

    async def aep_handle(self) -> AepResponse:
        """Handle AEP authentication."""
        aep_domain = self.domain

        if self._region_response.data.apiGatewayEndpoint is not None:
            aep_domain = self._region_response.data.apiGatewayEndpoint

        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=aep_domain,
        )
        client = Client(config)

        request = CommonParams(api_ver="1.0.0", language="en-US")
        logger.debug("client id %s", self._client_id)
        time_now = time.time()
        data_to_sign = {
            "appKey": self._app_key,
            "clientId": self._client_id,  # needs to be unique to device
            "deviceSn": self._device_sn,  # same here
            "timestamp": str(time_now),
        }

        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "authInfo": {
                    "clientId": self._client_id,
                    "sign": self.sign(data_to_sign),
                    "deviceSn": self._device_sn,
                    "timestamp": str(time_now),
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request("/app/aepauth/handle", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        response_body_str = response.body.decode("utf-8")

        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in getting mqtt credentials: " + response_body_dict["msg"])

        self._aep_response = AepResponse.from_dict(response_body_dict)

        logger.debug(response_body_dict)

        return self._aep_response

    async def connect(self) -> ConnectResponse:
        """Connect to the Aliyun Cloud IoT Gateway."""
        region_url = "sdk.openaccount.aliyun.com"
        time_now = time.time()
        async with ClientSession() as session:
            headers = {
                "host": region_url,
                "date": UtilClient.get_date_utcstring(),
                "x-ca-nonce": UtilClient.get_nonce(),
                "x-ca-key": self._app_key,
                "x-ca-signaturemethod": "HmacSHA256",
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
                "user-agent": UtilClient.get_user_agent(None),
            }

            _bodyParam = {
                "context": {
                    "sdkVersion": "3.4.2",
                    "platformName": "android",
                    "netType": "wifi",
                    "appKey": self._app_key,
                    "yunOSId": "",
                    "appVersion": APP_VERSION,
                    "utDid": self._utdid,
                    "appAuthToken": self._utdid,  # ???
                    "securityToken": self._utdid,  # ???
                },
                "config": {"version": 0, "lastModify": 0},
                "device": {
                    "model": "sdk_gphone_x86_arm",
                    "brand": "goldfish_x86",
                    "platformVersion": "30",
                },
            }

            # Get sign header
            dic = headers.copy()
            for key in MOVE_HEADERS:
                dic.pop(key, None)

            keys = sorted(dic.keys())
            sign_headers = ",".join(keys)
            header = "".join(f"{k}:{dic[k]}\n" for k in keys).strip()

            headers["x-ca-signature-headers"] = sign_headers
            string_to_sign = "POST\n{}\n\n{}\n{}\n{}\n/api/prd/connect.json?request={}".format(
                headers["accept"],
                headers["content-type"],
                headers["date"],
                header,
                json.dumps(_bodyParam, separators=(",", ":")),
            )

            hash_val = hmac.new(
                self._app_secret.encode("utf-8"),
                string_to_sign.encode("utf-8"),
                hashlib.sha256,
            ).digest()
            signature = base64.b64encode(hash_val).decode("utf-8")
            headers["x-ca-signature"] = signature

            async with session.post(
                f"https://{region_url}/api/prd/connect.json",
                headers=headers,
                params={"request": json.dumps(_bodyParam, separators=(",", ":"))},
            ) as resp:
                data = await resp.json()
                logger.debug(data)
                if resp.status == 200:
                    self._connect_response = ConnectResponse.from_dict(data)
                    return self._connect_response
                raise LoginException(data)

    async def login_by_oauth(self, country_code: str):
        """Login by OAuth."""
        auth_code = self.mammotion_http.login_info.authorization_code
        region_url = self._region_response.data.oaApiGatewayEndpoint

        async with ClientSession() as session:
            headers = {
                "host": region_url,
                "date": UtilClient.get_date_utcstring(),
                "x-ca-nonce": UtilClient.get_nonce(),
                "x-ca-key": self._app_key,
                "x-ca-signaturemethod": "HmacSHA256",
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded; charset=utf-8",
                "user-agent": UtilClient.get_user_agent(None),
                "vid": self._connect_response.data.vid,
            }

            _bodyParam = {
                "country": country_code,
                "authCode": auth_code,
                "oauthPlateform": 23,
                "oauthAppKey": self._app_key,
                "riskControlInfo": {
                    "appID": "com.agilexrobotics",
                    "appAuthToken": "",
                    "signType": "RSA",
                    "sdkVersion": "3.4.2",
                    "utdid": self._utdid,
                    "umidToken": self._utdid,
                    "deviceId": self._connect_response.data.data.device.data.deviceId,
                    "USE_OA_PWD_ENCRYPT": "true",
                    "USE_H5_NC": "true",
                },
            }

            # Get sign header
            dic = headers.copy()
            for key in MOVE_HEADERS:
                dic.pop(key, None)

            keys = sorted(dic.keys())
            sign_headers = ",".join(keys)
            header = "".join(f"{k}:{dic[k]}\n" for k in keys).strip()

            headers["x-ca-signature-headers"] = sign_headers
            string_to_sign = "POST\n{}\n\n{}\n{}\n{}\n/api/prd/loginbyoauth.json?{}".format(
                headers["accept"],
                headers["content-type"],
                headers["date"],
                header,
                f"loginByOauthRequest={json.dumps(_bodyParam, separators=(",", ":"))}",
            )

            hash_val = hmac.new(
                self._app_secret.encode("utf-8"),
                string_to_sign.encode("utf-8"),
                hashlib.sha256,
            ).digest()
            signature = base64.b64encode(hash_val).decode("utf-8")
            headers["x-ca-signature"] = signature
            async with session.post(
                f"https://{region_url}/api/prd/loginbyoauth.json",
                headers=headers,
                data={"loginByOauthRequest": json.dumps(_bodyParam, separators=(",", ":"))},
            ) as resp:
                data = await resp.json()
                logger.debug(data)
                if resp.status == 200:
                    self._login_by_oauth_response = LoginByOAuthResponse.from_dict(data)
                    return self._login_by_oauth_response
                raise LoginException(data)

    async def session_by_auth_code(self) -> SessionByAuthCodeResponse:
        """Create a session by auth code."""
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.4", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "request": {
                    "authCode": self._login_by_oauth_response.data.data.loginSuccessResult.sid,
                    "accountType": "OA_SESSION",
                    "appKey": self._app_key,
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request(
            "/account/createSessionByAuthCode",
            "https",
            "POST",
            None,
            body,
            RuntimeOptions(),
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        session_by_auth = SessionByAuthCodeResponse.from_dict(response_body_dict)

        if int(session_by_auth.code) != 200:
            raise Exception("Error in creating session: " + response_body_str)

        if session_by_auth.data.identityId is None:
            raise Exception("Error in creating session: " + response_body_str)

        self._session_by_authcode_response = session_by_auth
        self._iot_token_issued_at = int(time.time())

        return response.body

    async def sign_out(self) -> dict:
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.4", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "request": {
                    "refreshToken": self._session_by_authcode_response.data.refreshToken,
                    "identityId": self._session_by_authcode_response.data.identityId,
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        # possibly need to do this ourselves
        response = await client.async_do_request(
            "/iotx/account/invalidSession",
            "https",
            "POST",
            None,
            body,
            RuntimeOptions(),
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)
        return response_body_dict

    async def check_or_refresh_session(self) -> None:
        """Check or refresh the session."""
        logger.debug("Trying to refresh token")
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.4", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "request": {
                    "refreshToken": self._session_by_authcode_response.data.refreshToken,
                    "identityId": self._session_by_authcode_response.data.identityId,
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        # possibly need to do this ourselves
        response = await client.async_do_request(
            "/account/checkOrRefreshSession",
            "https",
            "POST",
            None,
            body,
            RuntimeOptions(),
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            logger.error(response_body_dict)
            await self.sign_out()
            raise CheckSessionException("Error check or refresh token: " + response_body_dict.__str__())

        if response_body_dict.get("code") == 2401:
            await self.sign_out()
            raise CheckSessionException("Error check or refresh token: " + response_body_dict.__str__())

        session = SessionByAuthCodeResponse.from_dict(response_body_dict)
        session_data = session.data

        if (
            session_data is None
            or session_data.identityId is None
            or session_data.refreshTokenExpire is None
            or session_data.iotToken is None
            or session_data.iotTokenExpire is None
            or session_data.refreshToken is None
        ):
            raise Exception("Error check or refresh token: Parameters not correct")

        self._session_by_authcode_response = session
        self._iot_token_issued_at = int(time.time())

    async def list_binding_by_account(self) -> ListingDevAccountResponse:
        """List bindings by account."""
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.8",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={"pageSize": 100, "pageNo": 1},
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request(
            "/uc/listBindingByAccount", "https", "POST", None, body, RuntimeOptions()
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in creating session: " + response_body_dict["message"])

        self._devices_by_account_response = ListingDevAccountResponse.from_dict(response_body_dict)
        return self._devices_by_account_response

    async def list_binding_by_dev(self, iot_id: str):
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.8",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={"pageSize": 100, "pageNo": 1, "iotId": iot_id},
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request("/uc/listBindingByDev", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in getting shared device list: " + response_body_dict["msg"])

        self._devices_by_account_response = ListingDevAccountResponse.from_dict(response_body_dict)
        return self._devices_by_account_response

    async def confirm_share(self, record_list: list[str]) -> bool:
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.7",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={"agree": 1, "recordIdList": record_list},
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request("/uc/confirmShare", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in accepting share: " + response_body_dict["msg"])

        return True

    async def get_shared_notice_list(self):
        ### status 0 accepted status -1 ready to be accepted 3 expired
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.9",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={"pageSize": 100, "pageNo": 1},
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request(
            "/uc/getShareNoticeList", "https", "POST", None, body, RuntimeOptions()
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in creating session: " + response_body_dict["msg"])

        self._devices_by_account_response = ListingDevAccountResponse.from_dict(response_body_dict)
        return self._devices_by_account_response

    async def send_cloud_command(self, iot_id: str, command: bytes) -> str:
        """Sends a cloud command to a specified IoT device.

        This function checks if the IoT token is expired and attempts to refresh it if
        possible. It then constructs a request using the provided command and sends it
        to the IoT device via an asynchronous HTTP POST request. The function handles
        various error codes and exceptions based on the response from the cloud
        service.

        Args:
            iot_id (str): The unique identifier of the IoT device.
            command (bytes): The command to be sent to the IoT device in binary format.

        Returns:
            str: A unique message ID for the sent command.

        """
        if command is None:
            raise Exception("Command is missing / None")

        """Check if iotToken is expired"""
        if self._iot_token_issued_at + self._session_by_authcode_response.data.iotTokenExpire <= (
            int(time.time()) + (5 * 3600)
        ):
            """Token expired - Try to refresh - Check if refreshToken is not expired"""
            if self._iot_token_issued_at + self._session_by_authcode_response.data.refreshTokenExpire > (
                int(time.time())
            ):
                await self.check_or_refresh_session()
            else:
                raise AuthRefreshException("Refresh token expired. Please re-login")

        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)
        # build request
        request = CommonParams(
            api_ver="1.0.5",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )

        # TODO move to using InvokeThingServiceRequest()

        message_id = str(uuid.uuid4())

        body = IoTApiRequest(
            id=message_id,
            params={
                "args": {"content": self.converter.printBase64Binary(command)},
                "identifier": "device_protobuf_sync_service",
                "iotId": f"{iot_id}",
            },
            request=request,
            version="1.0",
        )
        logger.debug(self.converter.printBase64Binary(command))
        # send request
        runtime_options = RuntimeOptions(autoretry=True, backoff_policy="yes")
        response = await client.async_do_request("/thing/service/invoke", "https", "POST", None, body, runtime_options)
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)
        logger.debug(iot_id)

        if response.status_code == 429:
            logger.debug("too many requests.")
            if self.message_delay > 8:
                raise TooManyRequestsException(response.status_message, iot_id)
            asyncio.get_event_loop().call_later(
                self.message_delay, lambda: asyncio.ensure_future(self.send_cloud_command(iot_id, command))
            )
            self.message_delay = self.message_delay * 2
            return message_id

        response_body_str = response.body.decode("utf-8")
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            logger.error(
                "Error in sending cloud command: %s - %s",
                str(response_body_dict.get("code")),
                str(response_body_dict.get("message")),
            )
            if response_body_dict.get("code") == 22000:
                logger.error(response.body)
                raise FailedRequestException(iot_id)
            if response_body_dict.get("code") == 20056:
                logger.debug("Gateway timeout.")
                raise GatewayTimeoutException(response_body_dict.get("code"), iot_id)

            if response_body_dict.get("code") == 29003:
                logger.debug(self._session_by_authcode_response.data.identityId)
                await self.sign_out()
                raise SetupException(response_body_dict.get("code"), iot_id)
            if response_body_dict.get("code") == 6205:
                raise DeviceOfflineException(response_body_dict.get("code"), iot_id)

            if response_body_dict.get("code") == 6205:
                raise CheckSessionException(response_body_dict.get("message"))

            if response_body_dict.get("code") == 460:
                logger.debug("iotToken expired, must re-login.")
                raise CheckSessionException(response_body_dict.get("message"))

        if self.message_delay != 1:
            self.message_delay = 1

        return message_id

    async def get_device_properties(self, iot_id: str) -> ThingPropertiesResponse:
        """List bindings by account."""
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.0",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "iotId": f"{iot_id}",
            },
            request=request,
            version="1.0",
        )

        # send request
        response = await client.async_do_request("/thing/properties/get", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = self.parse_json_response(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in getting properties: " + response_body_dict["msg"])

        return ThingPropertiesResponse.from_dict(response_body_dict)

    @property
    def devices_by_account_response(self):
        return self._devices_by_account_response

    def set_http(self, mammotion_http: MammotionHTTP) -> None:
        self.mammotion_http = mammotion_http

    @property
    def region_response(self) -> RegionResponse | None:
        return self._region_response

    @property
    def aep_response(self) -> AepResponse | None:
        return self._aep_response

    @property
    def session_by_authcode_response(self) -> SessionByAuthCodeResponse:
        return self._session_by_authcode_response

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def login_by_oauth_response(self) -> LoginByOAuthResponse | None:
        return self._login_by_oauth_response

    @property
    def connect_response(self) -> ConnectResponse | None:
        return self._connect_response



"""Device control of mammotion robots over bluetooth or MQTT."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

from aiohttp import ClientSession
from bleak import BLEDevice

from pymammotion import MammotionMQTT
from pymammotion.aliyun.cloud_gateway import CloudIOTGateway
from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model.device import MowingDevice
from pymammotion.data.model.enums import ConnectionPreference
from pymammotion.http.http import MammotionHTTP
from pymammotion.http.model.camera_stream import StreamSubscriptionResponse, VideoResourceResponse
from pymammotion.http.model.http import DeviceRecord, Response
from pymammotion.mammotion.devices.mammotion_cloud import MammotionCloud
from pymammotion.mammotion.devices.mammotion_mower_ble import MammotionMowerBLEDevice
from pymammotion.mammotion.devices.managers.managers import AbstractDeviceManager
from pymammotion.mammotion.devices.mower_manager import MammotionMowerDeviceManager
from pymammotion.mammotion.devices.rtk_manager import MammotionRTKDeviceManager
from pymammotion.mqtt import AliyunMQTT
from pymammotion.utility.device_type import DeviceType

# RTK imports - imported here for type hints, full import in add_cloud_devices
if TYPE_CHECKING:
    from pymammotion.mammotion.devices.rtk_ble import MammotionRTKBLEDevice

TIMEOUT_CLOUD_RESPONSE = 10

_LOGGER = logging.getLogger(__name__)


class MammotionDeviceManager:
    """Manage devices - both mowers and RTK."""

    def __init__(self) -> None:
        self.devices: dict[str, MammotionMowerDeviceManager] = {}
        self.rtk_devices: dict[str, MammotionRTKDeviceManager] = {}

    async def stop(self) -> None:
        for mower in self.devices.values():
            if cloud := mower.cloud:
                cloud.stop()
            if ble := mower.ble:
                await ble.stop()

    def _should_disconnect_mqtt(self, device_for_removal: AbstractDeviceManager) -> bool:
        """Check if MQTT connection should be disconnected.

        Returns True if no other devices share the same MQTT connection.
        """
        if not device_for_removal.cloud:
            return False

        mqtt_to_check = device_for_removal.cloud.mqtt

        # Check if any mower device shares this MQTT connection
        shared_devices: set[AbstractDeviceManager] = {
            device
            for device in self.devices.values()
            if device.cloud is not None and device.cloud.mqtt == mqtt_to_check
        }

        # Also check RTK devices for shared MQTT
        shared_devices.update(
            {
                device
                for device in self.rtk_devices.values()
                if device.cloud is not None and device.cloud.mqtt == mqtt_to_check
            }
        )

        return len(shared_devices) == 0

    def add_device(self, mammotion_device: MammotionMowerDeviceManager) -> None:
        """Add a mower device."""
        exists: MammotionMowerDeviceManager | None = self.devices.get(mammotion_device.name)
        if exists is None:
            self.devices[mammotion_device.name] = mammotion_device
            return
        if mammotion_device.cloud is not None:
            exists.replace_cloud(mammotion_device.cloud)
        if mammotion_device.ble:
            exists.replace_ble(mammotion_device.ble)

    def add_rtk_device(self, rtk_device: MammotionRTKDeviceManager) -> None:
        """Add an RTK device."""

        exists: MammotionRTKDeviceManager | None = self.rtk_devices.get(rtk_device.name)
        if exists is None:
            self.rtk_devices[rtk_device.name] = rtk_device
            return
        if rtk_device.cloud:
            exists.replace_cloud(rtk_device.cloud)
        if rtk_device.ble:
            exists.replace_ble(rtk_device.ble)

    def has_device(self, mammotion_device_name: str) -> bool:
        """Check if a mower device exists."""
        if self.devices.get(mammotion_device_name, None) is not None:
            return True
        return False

    def has_rtk_device(self, rtk_device_name: str) -> bool:
        """Check if an RTK device exists."""
        if self.rtk_devices.get(rtk_device_name, None) is not None:
            return True
        return False

    def get_device(self, mammotion_device_name: str) -> MammotionMowerDeviceManager:
        """Get a mower device."""
        return self.devices[mammotion_device_name]

    def get_device_by_iot_id(self, iot_id: str) -> MammotionMowerDeviceManager | None:
        """Get a mower device by IoT ID."""
        for device in self.devices.values():
            if device.iot_id == iot_id:
                return device
        return None

    def get_rtk_device(self, rtk_device_name: str) -> MammotionRTKDeviceManager:
        """Get an RTK device."""
        return self.rtk_devices[rtk_device_name]

    async def remove_device(self, name: str) -> None:
        """Remove a mower device."""
        if self.devices.get(name):
            device_for_removal = self.devices.pop(name)
            loop = asyncio.get_running_loop()

            if device_for_removal.cloud:
                if self._should_disconnect_mqtt(device_for_removal):
                    await loop.run_in_executor(None, device_for_removal.cloud.mqtt.disconnect)
                    device_for_removal.cloud.stop()

            if device_for_removal.ble:
                await device_for_removal.ble.stop()

            del device_for_removal

    async def remove_rtk_device(self, name: str) -> None:
        """Remove an RTK device."""
        if self.rtk_devices.get(name):
            device_for_removal = self.rtk_devices.pop(name)
            loop = asyncio.get_running_loop()

            if device_for_removal.cloud:
                if self._should_disconnect_mqtt(device_for_removal):
                    await loop.run_in_executor(None, device_for_removal.cloud.mqtt.disconnect)
                    device_for_removal.cloud.stop()

            if device_for_removal.ble:
                await device_for_removal.ble.stop()

            del device_for_removal


class Mammotion:
    """Represents a Mammotion account and its devices."""

    device_manager = MammotionDeviceManager()

    _instance: Mammotion | None = None

    def __new__(cls) -> Mammotion:
        """Create a singleton."""
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, session: ClientSession | None = None) -> None:
        """Initialize MammotionDevice."""
        self._login_lock = asyncio.Lock()
        self._session: ClientSession | None = session
        self.mqtt_list: dict[str, MammotionCloud] = {}

    async def login_and_initiate_cloud(self, account: str, password: str, force: bool = False) -> None:
        async with self._login_lock:
            exists_aliyun: MammotionCloud | None = self.mqtt_list.get(f"{account}_aliyun")
            exists_mammotion: MammotionCloud | None = self.mqtt_list.get(f"{account}_mammotion")
            if (not exists_aliyun and not exists_mammotion) or force:
                cloud_client = await self.login(account, password)
                await self.initiate_cloud_connection(account, cloud_client)

    async def refresh_login(self, account: str) -> None:
        """Refresh login."""
        async with self._login_lock:
            exists_aliyun: MammotionCloud | None = self.mqtt_list.get(f"{account}_aliyun")
            exists_mammotion: MammotionCloud | None = self.mqtt_list.get(f"{account}_mammotion")

            if not exists_aliyun and not exists_mammotion:
                return
            mammotion_http = (
                exists_aliyun.cloud_client.mammotion_http
                if exists_aliyun
                else exists_mammotion.cloud_client.mammotion_http
            )

            await mammotion_http.refresh_login()

            if len(mammotion_http.device_records.records) != 0:
                await mammotion_http.get_mqtt_credentials()

            if exists_aliyun and exists_aliyun.is_connected():
                exists_aliyun.disconnect()
                await self.connect_iot(exists_aliyun.cloud_client)

            if exists_mammotion and exists_mammotion.is_connected():
                exists_mammotion.disconnect()

            if exists_aliyun and not exists_aliyun.is_connected():
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(None, exists_aliyun.connect_async)
            if exists_mammotion and not exists_mammotion.is_connected():
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(None, exists_mammotion.connect_async)

    @staticmethod
    def shim_cloud_devices(devices: list[DeviceRecord]) -> list[Device]:
        device_list: list[Device] = []
        for device in devices:
            device_list.append(
                Device(
                    gmt_modified=0,
                    product_name="",
                    status=0,
                    net_type="NET_WIFI",
                    is_edge_gateway=False,
                    category_name="",
                    owned=1,
                    identity_alias="UNKNOW",
                    thing_type="DEVICE",
                    identity_id=device.identity_id,
                    device_name=device.device_name,
                    product_key=device.product_key,
                    iot_id=device.iot_id,
                    bind_time=device.bind_time,
                    node_type="DEVICE",
                    category_key="LawnMower",
                )
            )

        return device_list

    async def initiate_ble_connection(self, devices: dict[str, BLEDevice], cloud_devices: list[Device]) -> None:
        """Initiate BLE connection."""
        for device in cloud_devices:
            if ble_device := devices.get(device.device_name):
                if device.device_name.startswith(("Luba-", "Yuka-")):
                    if not self.device_manager.has_device(device.device_name):
                        self.device_manager.add_device(
                            MammotionMowerDeviceManager(
                                name=device.device_name,
                                iot_id=device.iot_id,
                                cloud_device=device,
                                ble_device=ble_device,
                                preference=ConnectionPreference.BLUETOOTH,
                                cloud_client=CloudIOTGateway(MammotionHTTP(session=self._session)),
                            )
                        )
                    else:
                        self.device_manager.get_device(device.device_name).add_ble(ble_device)
                if device.device_name.startswith(("RTK", "RBS")):
                    if not self.device_manager.has_rtk_device(device.device_name):
                        self.device_manager.add_rtk_device(
                            MammotionRTKDeviceManager(
                                name=device.device_name,
                                iot_id=device.iot_id,
                                cloud_device=device,
                                ble_device=ble_device,
                                preference=ConnectionPreference.BLUETOOTH,
                                cloud_client=CloudIOTGateway(MammotionHTTP(session=self._session)),
                            )
                        )
                    else:
                        self.device_manager.get_rtk_device(device.device_name).add_ble(ble_device)

    async def initiate_cloud_connection(self, account: str, cloud_client: CloudIOTGateway) -> None:
        """Initiate cloud connection."""
        loop = asyncio.get_running_loop()

        mammotion_http = cloud_client.mammotion_http

        if mqtt := self.mqtt_list.get(f"{account}_aliyun"):
            if mqtt.is_connected():
                await loop.run_in_executor(None, mqtt.disconnect)

        if mqtt := self.mqtt_list.get(f"{account}_mammotion"):
            if mqtt.is_connected():
                await loop.run_in_executor(None, mqtt.disconnect)

        if len(cloud_client.devices_by_account_response.data.data) != 0:
            mammotion_cloud = MammotionCloud(
                AliyunMQTT(
                    region_id=cloud_client.region_response.data.regionId,
                    product_key=cloud_client.aep_response.data.productKey,
                    device_name=cloud_client.aep_response.data.deviceName,
                    device_secret=cloud_client.aep_response.data.deviceSecret,
                    iot_token=cloud_client.session_by_authcode_response.data.iotToken,
                    client_id=cloud_client.client_id,
                    cloud_client=cloud_client,
                ),
                cloud_client,
            )
            self.mqtt_list[f"{account}_aliyun"] = mammotion_cloud
            self.add_cloud_devices(mammotion_cloud)

            await loop.run_in_executor(None, self.mqtt_list[f"{account}_aliyun"].connect_async)
        if len(mammotion_http.device_records.records) != 0 and mammotion_http.mqtt_credentials is not None:
            mammotion_cloud = MammotionCloud(
                MammotionMQTT(
                    records=mammotion_http.device_records.records,
                    mammotion_http=mammotion_http,
                    mqtt_connection=mammotion_http.mqtt_credentials,
                ),
                cloud_client,
            )
            self.mqtt_list[f"{account}_mammotion"] = mammotion_cloud
            self.add_mammotion_devices(mammotion_cloud, mammotion_http.device_records.records)

            await loop.run_in_executor(None, self.mqtt_list[f"{account}_mammotion"].connect_async)

    def add_mammotion_devices(self, mqtt_client: MammotionCloud, devices: list[DeviceRecord]) -> None:
        """Add devices from mammotion cloud."""
        for device in devices:
            if device.device_name.startswith(("Luba-", "Yuka-")):
                has_device = self.device_manager.has_device(device.device_name)
                if has_device:
                    mower_device = self.device_manager.get_device(device.device_name)
                    if mower_device.cloud is None:
                        mower_device.add_cloud(mqtt=mqtt_client)
                    else:
                        mower_device.replace_mqtt(mqtt_client)

                else:
                    cloud_device_shim = Device(
                        gmt_modified=0,
                        product_name="",
                        status=0,
                        net_type="NET_WIFI",
                        is_edge_gateway=False,
                        category_name="",
                        owned=1,
                        identity_alias="UNKNOW",
                        thing_type="DEVICE",
                        identity_id=device.identity_id,
                        device_name=device.device_name,
                        product_key=device.product_key,
                        iot_id=device.iot_id,
                        bind_time=device.bind_time,
                        node_type="DEVICE",
                        category_key="LawnMower",
                    )

                    mixed_device = MammotionMowerDeviceManager(
                        name=device.device_name,
                        iot_id=device.iot_id,
                        cloud_client=mqtt_client.cloud_client,
                        cloud_device=cloud_device_shim,
                        mqtt=mqtt_client,
                        preference=ConnectionPreference.WIFI,
                    )
                    mixed_device.state.mower_state.product_key = device.product_key
                    self.device_manager.add_device(mixed_device)

    def add_cloud_devices(self, mqtt_client: MammotionCloud) -> None:
        """Add devices from cloud - both mowers and RTK."""
        from pymammotion.mammotion.devices.rtk_manager import MammotionRTKDeviceManager

        for device in mqtt_client.cloud_client.devices_by_account_response.data.data:
            # Handle mower devices (Luba, Yuka)
            if device.device_name.startswith(("Luba-", "Yuka-")):
                has_device = self.device_manager.has_device(device.device_name)
                if not has_device:
                    mixed_device = MammotionMowerDeviceManager(
                        name=device.device_name,
                        iot_id=device.iot_id,
                        cloud_client=mqtt_client.cloud_client,
                        cloud_device=device,
                        mqtt=mqtt_client,
                        preference=ConnectionPreference.WIFI,
                    )
                    mixed_device.state.mower_state.product_key = device.product_key
                    mixed_device.state.mower_state.model = (
                        device.product_name if device.product_model is None else device.product_model
                    )
                    self.device_manager.add_device(mixed_device)
                else:
                    mower_device = self.device_manager.get_device(device.device_name)
                    if mower_device.cloud is None:
                        mower_device.add_cloud(mqtt=mqtt_client)
                    else:
                        mower_device.replace_mqtt(mqtt_client)

            # Handle RTK devices
            elif device.device_name.startswith(("RTK", "RBS")):
                has_rtk_device = self.device_manager.has_rtk_device(device.device_name)
                if not has_rtk_device:
                    rtk_device = MammotionRTKDeviceManager(
                        name=device.device_name,
                        iot_id=device.iot_id,
                        cloud_client=mqtt_client.cloud_client,
                        cloud_device=device,
                        mqtt=mqtt_client,
                        preference=ConnectionPreference.WIFI,
                    )
                    self.device_manager.add_rtk_device(rtk_device)
                else:
                    rtk_device = self.device_manager.get_rtk_device(device.device_name)
                    if rtk_device.cloud is None:
                        rtk_device.add_cloud(mqtt=mqtt_client)
                    else:
                        rtk_device.replace_mqtt(mqtt_client)

    def set_disconnect_strategy(self, *, disconnect: bool) -> None:
        """Set disconnect strategy for all BLE devices (mowers and RTK)."""
        for device in self.device_manager.devices.values():
            if device.ble is not None:
                ble_device: MammotionMowerBLEDevice = device.ble
                ble_device.set_disconnect_strategy(disconnect=disconnect)

        for rtk_device in self.device_manager.rtk_devices.values():
            if rtk_device.ble is not None:
                ble_rtk_device: MammotionRTKBLEDevice = rtk_device.ble
                ble_rtk_device.set_disconnect_strategy(disconnect=disconnect)

    async def login(self, account: str, password: str) -> CloudIOTGateway:
        """Login to mammotion cloud."""
        mammotion_http = MammotionHTTP(session=self._session)
        await mammotion_http.login_v2(account, password)
        await mammotion_http.get_user_device_page()
        device_list = await mammotion_http.get_user_device_list()
        _LOGGER.debug("device_list: %s", device_list)
        await mammotion_http.get_mqtt_credentials()
        cloud_client = CloudIOTGateway(mammotion_http)
        await self.connect_iot(cloud_client)
        return cloud_client

    async def stop(self) -> None:
        await self.device_manager.stop()
        for mqtt in self.mqtt_list.values():
            if mqtt.is_connected():
                mqtt.disconnect()

    @staticmethod
    async def connect_iot(cloud_client: CloudIOTGateway) -> None:
        """Connect to aliyun cloud and fetch device info."""
        mammotion_http = cloud_client.mammotion_http
        country_code = mammotion_http.login_info.userInformation.domainAbbreviation
        if cloud_client.region_response is None:
            await cloud_client.get_region(country_code)
        await cloud_client.connect()
        await cloud_client.login_by_oauth(country_code)
        await cloud_client.aep_handle()
        await cloud_client.session_by_auth_code()
        await cloud_client.list_binding_by_account()

    async def remove_device(self, name: str) -> None:
        """Remove a mower device."""
        await self.device_manager.remove_device(name)

    async def remove_rtk_device(self, name: str) -> None:
        """Remove an RTK device."""
        await self.device_manager.remove_rtk_device(name)

    def get_device_by_name(self, name: str) -> MammotionMowerDeviceManager:
        """Get a mower device by name."""
        return self.device_manager.get_device(name)

    def get_rtk_device_by_name(self, name: str) -> MammotionRTKDeviceManager:
        """Get an RTK device by name."""
        return self.device_manager.get_rtk_device(name)

    def get_or_create_device_by_name(
        self, device: Device, mqtt_client: MammotionCloud | None, ble_device: BLEDevice | None
    ) -> MammotionMowerDeviceManager:
        """Get or create a mower device by name."""
        if self.device_manager.has_device(device.device_name):
            return self.device_manager.get_device(device.device_name)
        mow_device = MammotionMowerDeviceManager(
            name=device.device_name,
            iot_id=device.iot_id,
            cloud_client=mqtt_client.cloud_client
            if mqtt_client
            else CloudIOTGateway(MammotionHTTP(session=self._session)),
            mqtt=mqtt_client,
            cloud_device=device,
            ble_device=ble_device,
            preference=ConnectionPreference.WIFI if mqtt_client else ConnectionPreference.BLUETOOTH,
        )
        self.device_manager.add_device(mow_device)
        return mow_device

    async def send_command(self, name: str, key: str):
        """Send a command to the device."""
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH and device.ble:
                return await device.ble.command(key)
            if device.preference is ConnectionPreference.WIFI and device.cloud:
                return await device.cloud.command(key)
            # TODO work with both with EITHER
        return None

    async def send_command_with_args(self, name: str, key: str, **kwargs: Any):
        """Send a command with args to the device."""
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH and device.ble:
                return await device.ble.command(key, **kwargs)
            if device.preference is ConnectionPreference.WIFI and device.cloud:
                return await device.cloud.command(key, **kwargs)
            # TODO work with both with EITHER
        return None

    async def start_map_sync(self, name: str) -> None:
        """Start map sync."""
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH and device.ble:
                return await device.ble.start_map_sync()
            if device.preference is ConnectionPreference.WIFI and device.cloud:
                return await device.cloud.start_map_sync()
            # TODO work with both with EITHER
        return None

    async def start_schedule_sync(self, name: str) -> None:
        """Start map sync."""
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH and device.ble:
                return await device.ble.start_schedule_sync()
            if device.preference is ConnectionPreference.WIFI and device.cloud:
                return await device.cloud.start_schedule_sync()
            # TODO work with both with EITHER
        return None

    async def get_stream_subscription(self, name: str, iot_id: str) -> Response[StreamSubscriptionResponse] | Any:
        """Get stream subscription."""
        device = self.get_device_by_name(name)
        _stream_response = await device.mammotion_http.get_stream_subscription(
            iot_id, DeviceType.is_yuka(name) and not DeviceType.is_yuka_mini(name)
        )
        _LOGGER.debug(_stream_response)
        return _stream_response

    async def get_video_resource(self, name: str, iot_id: str) -> Response[VideoResourceResponse] | None:
        """Get video resource."""
        device = self.get_device_by_name(name)

        if DeviceType.is_mini_or_x_series(name):
            _video_resource_response = await device.mammotion_http.get_video_resource(iot_id)
            _LOGGER.debug(_video_resource_response)
            return _video_resource_response
        return None

    def mower(self, name: str) -> MowingDevice | None:
        """Get a mower device by name."""
        device = self.get_device_by_name(name)
        if device:
            return device.state
        return None



"""Mower-specific device class with map synchronization callbacks."""

from abc import ABC
import logging

from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model import RegionData
from pymammotion.data.mower_state_manager import MowerStateManager
from pymammotion.mammotion.devices.base import MammotionBaseDevice
from pymammotion.proto import NavGetCommDataAck, NavGetHashListAck, NavPlanJobSet, SvgMessageAckT
from pymammotion.utility.device_type import DeviceType

_LOGGER = logging.getLogger(__name__)


def find_next_integer(lst: list[int], current_hash: int) -> int | None:
    """Find the next integer in a list after the current hash."""
    try:
        current_index = lst.index(current_hash)
        if current_index + 1 < len(lst):
            return lst[current_index + 1]
        else:
            return None
    except ValueError:
        return None


class MammotionMowerDevice(MammotionBaseDevice, ABC):
    """Mower device with map synchronization support."""

    def __init__(self, state_manager: MowerStateManager, cloud_device: Device) -> None:
        """Initialize MammotionMowerDevice."""
        super().__init__(state_manager, cloud_device)

    async def datahash_response(self, hash_ack: NavGetHashListAck) -> None:
        """Handle datahash responses for root level hashs."""
        current_frame = hash_ack.current_frame

        missing_frames = self.mower.map.missing_root_hash_frame(hash_ack)
        if len(missing_frames) == 0:
            if len(self.mower.map.missing_hashlist(hash_ack.sub_cmd)) > 0:
                data_hash = self.mower.map.missing_hashlist(hash_ack.sub_cmd).pop(0)
                await self.queue_command("synchronize_hash_data", hash_num=data_hash)
            return

        if current_frame != missing_frames[0] - 1:
            current_frame = missing_frames[0] - 1
        await self.queue_command("get_hash_response", total_frame=hash_ack.total_frame, current_frame=current_frame)

    async def commdata_response(self, common_data: NavGetCommDataAck | SvgMessageAckT) -> None:
        """Handle common data responses."""
        total_frame = common_data.total_frame
        current_frame = common_data.current_frame

        missing_frames = self.mower.map.missing_frame(common_data)
        if len(missing_frames) == 0:
            # get next in hash ack list
            data_hash = (
                self.mower.map.missing_hashlist(common_data.sub_cmd).pop(0)
                if len(self.mower.map.missing_hashlist(common_data.sub_cmd)) > 0
                else None
            )
            if data_hash is None:
                return

            await self.queue_command("synchronize_hash_data", hash_num=data_hash)
        else:
            if current_frame != missing_frames[0] - 1:
                current_frame = missing_frames[0] - 1

            region_data = RegionData()
            region_data.hash = common_data.data_hash if isinstance(common_data, SvgMessageAckT) else common_data.hash
            region_data.action = common_data.action if isinstance(common_data, NavGetCommDataAck) else 0
            region_data.type = common_data.type
            region_data.sub_cmd = common_data.sub_cmd
            region_data.total_frame = total_frame
            region_data.current_frame = current_frame
            await self.queue_command("get_regional_data", regional_data=region_data)

    async def plan_callback(self, plan: NavPlanJobSet) -> None:
        """Handle plan job responses."""
        if plan.plan_index < plan.total_plan_num - 1:
            index = plan.plan_index + 1
            await self.queue_command("read_plan", sub_cmd=2, plan_index=index)

    async def start_schedule_sync(self) -> None:
        """Start sync of schedule data."""
        if len(self.mower.map.plan) == 0 or list(self.mower.map.plan.values())[0].total_plan_num != len(
            self.mower.map.plan
        ):
            await self.queue_command("read_plan", sub_cmd=2, plan_index=0)

    async def start_map_sync(self) -> None:
        """Start sync of map data."""
        if location := next((loc for loc in self.mower.report_data.locations if loc.pos_type == 5), None):
            self.mower.map.update_hash_lists(self.mower.map.hashlist, location.bol_hash)

        await self.queue_command("send_todev_ble_sync", sync_type=3)

        # TODO correctly check if area names exist for a zone.
        if self._cloud_device and len(self.mower.map.area_name) == 0 and not DeviceType.is_luba1(self.mower.name):
            await self.queue_command("get_area_name_list", device_id=self._cloud_device.iot_id)

        if len(self.mower.map.root_hash_lists) == 0 or len(self.mower.map.missing_hashlist()) > 0:
            await self.queue_command("get_all_boundary_hash_list", sub_cmd=0)

        for hash_id, frame in list(self.mower.map.area.items()):
            missing_frames = self.mower.map.find_missing_frames(frame)
            if len(missing_frames) > 0:
                del self.mower.map.area[hash_id]

        for hash_id, frame in list(self.mower.map.path.items()):
            missing_frames = self.mower.map.find_missing_frames(frame)
            if len(missing_frames) > 0:
                del self.mower.map.path[hash_id]

        for hash_id, frame in list(self.mower.map.obstacle.items()):
            missing_frames = self.mower.map.find_missing_frames(frame)
            if len(missing_frames) > 0:
                del self.mower.map.obstacle[hash_id]



from dataclasses import dataclass
from typing import Any

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class TopicProperty(DataClassORJSONMixin):
    """TopicProperty."""

    id: str = ""
    method: str = ""
    params: dict[str, Any] | None = None
    sys: dict[str, Any] | None = None
    time: int = 0
    version: str = ""


@dataclass
class TopicDeviceStatus(DataClassORJSONMixin):
    """TopicDeviceStatus."""

    gmt_create: int = 0
    action: str = ""
    product_key: str = ""
    device_name: str = ""
    iot_id: str = ""


class TopicUtils:
    """Utility helpers ported from the Java TopicUtils."""

    @staticmethod
    def _split_topic(topic: str) -> list[str]:
        if topic is None:
            raise ValueError("topic must not be None")
        # preserve empty segments (leading/trailing slashes)
        return topic.split("/")

    @staticmethod
    def get_device_name(topic: str) -> str:
        parts = TopicUtils._split_topic(topic)
        # original code expects the device name at index 3 when topic is like:
        # /sys/{productKey}/{deviceName}/...
        try:
            return parts[3]
        except IndexError:
            return ""

    @staticmethod
    def get_identifier(topic: str) -> str:
        if "property" in (topic or ""):
            return ""
        parts = TopicUtils._split_topic(topic)
        if len(parts) < 2:
            return ""
        # second-last element (may be empty if trailing slash)
        return parts[-2] or ""

    @staticmethod
    def get_method(topic: str) -> str:
        """Get the method from a topic."""
        parts = TopicUtils._split_topic(topic)
        if len(parts) >= 2:
            return f"/{parts[-2]}/{parts[-1]}"
        return ""



"""RTK device with Bluetooth LE connectivity."""

import asyncio
import logging
from typing import Any
from uuid import UUID

from bleak import BleakGATTCharacteristic, BLEDevice
from bleak_retry_connector import BleakClientWithServiceCache

from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.bluetooth import BleMessage
from pymammotion.data.model.device import RTKDevice
from pymammotion.mammotion.commands.mammotion_command import MammotionCommand
from pymammotion.mammotion.devices.rtk_device import MammotionRTKDevice

_LOGGER = logging.getLogger(__name__)


class MammotionRTKBLEDevice(MammotionRTKDevice):
    """RTK device with BLE connectivity - simpler than mowers, no map sync."""

    def __init__(
        self, cloud_device: Device, rtk_state: RTKDevice, device: BLEDevice, interface: int = 0, **kwargs: Any
    ) -> None:
        """Initialize MammotionRTKBLEDevice."""
        super().__init__(cloud_device, rtk_state)
        self.command_sent_time = 0
        self._disconnect_strategy = True
        self._interface = f"hci{interface}"
        self.ble_device = device
        self._client: BleakClientWithServiceCache | None = None
        self._read_char: BleakGATTCharacteristic | int | str | UUID = 0
        self._write_char: BleakGATTCharacteristic | int | str | UUID = 0
        self._disconnect_timer: asyncio.TimerHandle | None = None
        self._message: BleMessage | None = None
        self._commands: MammotionCommand = MammotionCommand(device.name or "", 1)
        self.command_queue = asyncio.Queue()
        self._expected_disconnect = False
        self._connect_lock = asyncio.Lock()
        self._operation_lock = asyncio.Lock()
        self._key: str | None = None
        loop = asyncio.get_event_loop()
        loop.create_task(self.process_queue())

    def __del__(self) -> None:
        """Cleanup."""
        if self._disconnect_timer:
            self._disconnect_timer.cancel()

    @property
    def client(self) -> BleakClientWithServiceCache:
        """Return the BLE client."""
        return self._client

    def set_disconnect_strategy(self, *, disconnect: bool) -> None:
        """Set disconnect strategy."""
        self._disconnect_strategy = disconnect

    async def process_queue(self) -> None:
        """Process queued commands - simplified for RTK."""
        while True:
            key, kwargs = await self.command_queue.get()
            try:
                _LOGGER.debug("Processing RTK BLE command: %s", key)
                command_bytes = getattr(self._commands, key)(**kwargs)
                # Send command via BLE (implementation depends on BLE infrastructure)
                # For now, this is a placeholder
                _LOGGER.debug("RTK BLE command sent: %s", key)
            except Exception as ex:
                _LOGGER.exception("Error processing RTK BLE command: %s", ex)
            finally:
                self.command_queue.task_done()

    async def queue_command(self, key: str, **kwargs: Any) -> None:
        """Queue a command to the RTK device."""
        await self.command_queue.put((key, kwargs))

    async def command(self, key: str, **kwargs):
        """Send a command to the RTK device."""
        return await self.queue_command(key, **kwargs)

    async def _ble_sync(self) -> None:
        """RTK devices don't use BLE sync in the same way as mowers."""

    async def stop(self) -> None:
        """Stop everything ready for destroying."""
        if self._client is not None and self._client.is_connected:
            await self._client.disconnect()



from dataclasses import dataclass, field
import logging

logger = logging.getLogger(__name__)


@dataclass
class GenerateRouteInformation:
    """Creates a model for generating route information and mowing plan before starting a job."""

    one_hashs: list[int] = field(default_factory=list)
    job_mode: int = 4  # taskMode
    job_version: int = 0
    job_id: int = 0
    speed: float = 0.3
    ultra_wave: int = 2  # touch no touch etc
    channel_mode: int = 0  # line mode is grid single double or single2
    channel_width: int = 25
    rain_tactics: int = 0
    blade_height: int = 0
    path_order: str = ""
    toward: int = 0  # is just angle
    toward_included_angle: int = 0
    toward_mode: int = 0  # angle type relative etc
    edge_mode: int = 1  # border laps
    obstacle_laps: int = 1



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MctlPept(_message.Message):
    __slots__ = ["perception_local_map_visualization", "perception_obstacles_visualization", "perception_universal_buff"]
    PERCEPTION_LOCAL_MAP_VISUALIZATION_FIELD_NUMBER: _ClassVar[int]
    PERCEPTION_OBSTACLES_VISUALIZATION_FIELD_NUMBER: _ClassVar[int]
    PERCEPTION_UNIVERSAL_BUFF_FIELD_NUMBER: _ClassVar[int]
    perception_local_map_visualization: perception_local_map_visualization_t
    perception_obstacles_visualization: perception_obstacles_visualization_t
    perception_universal_buff: perception_universal_buff_t
    def __init__(self, perception_obstacles_visualization: _Optional[_Union[perception_obstacles_visualization_t, _Mapping]] = ..., perception_universal_buff: _Optional[_Union[perception_universal_buff_t, _Mapping]] = ..., perception_local_map_visualization: _Optional[_Union[perception_local_map_visualization_t, _Mapping]] = ...) -> None: ...

class perception_local_map_visualization_t(_message.Message):
    __slots__ = ["map"]
    MAP_FIELD_NUMBER: _ClassVar[int]
    map: bytes
    def __init__(self, map: _Optional[bytes] = ...) -> None: ...

class perception_obstacles_t(_message.Message):
    __slots__ = ["label", "num", "points_x", "points_y"]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    NUM_FIELD_NUMBER: _ClassVar[int]
    POINTS_X_FIELD_NUMBER: _ClassVar[int]
    POINTS_Y_FIELD_NUMBER: _ClassVar[int]
    label: int
    num: int
    points_x: _containers.RepeatedScalarFieldContainer[int]
    points_y: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, label: _Optional[int] = ..., num: _Optional[int] = ..., points_x: _Optional[_Iterable[int]] = ..., points_y: _Optional[_Iterable[int]] = ...) -> None: ...

class perception_obstacles_visualization_t(_message.Message):
    __slots__ = ["num", "obstacles", "scale", "status", "timestamp"]
    NUM_FIELD_NUMBER: _ClassVar[int]
    OBSTACLES_FIELD_NUMBER: _ClassVar[int]
    SCALE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    num: int
    obstacles: _containers.RepeatedCompositeFieldContainer[perception_obstacles_t]
    scale: float
    status: int
    timestamp: float
    def __init__(self, status: _Optional[int] = ..., num: _Optional[int] = ..., obstacles: _Optional[_Iterable[_Union[perception_obstacles_t, _Mapping]]] = ..., timestamp: _Optional[float] = ..., scale: _Optional[float] = ...) -> None: ...

class perception_universal_buff_t(_message.Message):
    __slots__ = ["perception_len", "perception_type", "universal_buff"]
    PERCEPTION_LEN_FIELD_NUMBER: _ClassVar[int]
    PERCEPTION_TYPE_FIELD_NUMBER: _ClassVar[int]
    UNIVERSAL_BUFF_FIELD_NUMBER: _ClassVar[int]
    perception_len: int
    perception_type: int
    universal_buff: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, perception_type: _Optional[int] = ..., perception_len: _Optional[int] = ..., universal_buff: _Optional[_Iterable[int]] = ...) -> None: ...



from dataclasses import dataclass, field

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class SideLight(DataClassORJSONMixin):
    operate: int = 0
    enable: int = 0
    start_hour: int = 0
    start_min: int = 0
    end_hour: int = 0
    end_min: int = 0
    action: int = 0


@dataclass
class DeviceNonWorkingHours(DataClassORJSONMixin):
    sub_cmd: int = 0
    start_time: str = ""
    end_time: str = ""


@dataclass
class LampInfo(DataClassORJSONMixin):
    lamp_bright: int = 0
    manual_light: bool = False
    night_light: bool = False


@dataclass
class MowerInfo(DataClassORJSONMixin):
    blade_status: bool = False
    rain_detection: bool = False
    traversal_mode: int = 0
    turning_mode: int = 0
    blade_mode: int = 0
    blade_rpm: int = 0
    side_led: SideLight = field(default_factory=SideLight)
    collector_installation_status: bool = False
    model: str = ""
    swversion: str = ""
    product_key: str = ""
    model_id: str = ""
    sub_model_id: str = ""
    ble_mac: str = ""
    wifi_mac: str = ""
    lamp_info: LampInfo = field(default_factory=LampInfo)


@dataclass
class DeviceFirmwares(DataClassORJSONMixin):
    device_version: str = ""
    left_motor_driver: str = ""
    lora_version: str = ""
    main_controller: str = ""
    model_name: str = ""
    right_motor_driver: str = ""
    rtk_rover_station: str = ""
    rtk_version: str = ""



from dataclasses import dataclass, field

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class WorkTaskEvent(DataClassORJSONMixin):
    hash_area_map: dict[int, int] = field(default_factory=dict)
    ids: list[int] = field(default_factory=list)


@dataclass
class Events(DataClassORJSONMixin):
    work_tasks_event: WorkTaskEvent = field(default_factory=WorkTaskEvent)
